//******************************************************************************
// cluster.java:	Applet
//
//******************************************************************************
import java.applet.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import mylib.*;
import treelib.ELTree;
import RngPack.Ranecu;

abstract class distMeasure 
{
  // a class for encapsulating measures of distance between vectors of
  // items
  public distMeasure ()
    {};

  final static int EUCLIDEAN = 0;
  final static int BRAY_CURTIS = 1;
  final static int CANBERRA = 2;
  final static int JACCARD = 3;
  final static int SORENSEN = 4;
  final static int SIMPLE_MATCHING = 5;
  final static int BARONI_URBANI_BUSER = 6;
  final static int MEAN_CENSORED_EUCLIDEAN = 7;

  final static String distNames[] = {"Euclidean Distance",
				     "Bray-Curtis Distance",
				     "Canberra Distance",
				     "Jaccard Similarity",
				     "Sorensen Similarity",
				     "Simple Matching Coefficient",
				     "Baroni-Urbani Buser Coefficient",
				     "Mean Censored Euclidean Distance"};

  public static final String getName (int i)
    {
      return (i < distNames.length) ? distNames[i] : "";
    }

  public static final int getNumMeasures ()
    {
      return distNames.length;
    }

  public abstract double distance (Object o1, Object o2);

  public static distMeasure getDistMeasure (int which)
    {
      distMeasure dm;
		
      switch (which)
	{
	case BRAY_CURTIS:
	  dm = new distBrayCurtis ();
	  break;
	case CANBERRA:
	  dm = new distCanberra ();
	  break;
	case JACCARD:
	  dm = new distJaccard ();
	  break;
	case SORENSEN:
	  dm = new distSorensen ();
	  break;
	case SIMPLE_MATCHING:
	  dm = new distSimpleMatching ();
	  break;
	case BARONI_URBANI_BUSER:
	  dm = new distBaroniUrbani ();
	  break;
	case MEAN_CENSORED_EUCLIDEAN:
	  dm = new distMeanCensoredEuclidean ();
	  break;
	default:
	  dm = new distEuclidean ();
	}
      return dm;
    }
}

abstract class distFromBinarySim extends distMeasure
{
  // a class for distances derived from binary similarity measures,
  // namely those which involve a, b, c, and d, where:
  //		a is the number of items present in both v1, v2
  //		b, c are the number of items present only in v1, v2, respectively
  //		d is the number of items absent in both v1, v2
	
  // all measures use the same method to get a, b, c, and d

  int a, b, c, d;

  public final void getCounts (float v1[], float v2[])
    {
      int x[] = new int[4];
      for (int i=0; i < v1.length; ++i)
	++ x[ ((v1[i] > 0) ? 0 : 2) + ((v2[i] > 0) ? 0 : 1)];

      a=x[0]; b=x[1]; c=x[2]; d=x[3];
    }

  public final void getCounts (SparseVector v1, SparseVector v2)
    {
      // make sure the vectors are sorted
      v1.sort ();
      v2.sort ();

      int ind1[] = v1.index, ind2[] = v2.index;
      float f1[] = v1.value, f2[] = v2.value;

      int i1 = 0, i2 = 0;
      int x[] = new int[4];

      // while there is some of both lists left

      while (i1 < f1.length && i2 < f2.length)
	{
	  if (ind1[i1] == ind2[i2])
	    // a member of both lists, increment appropriate x
	    // (the member might have value zero even though it is listed)
	    ++ x[ ((f1[i1++] > 0) ? 0 : 2) + ((f2[i2++] > 0) ? 0 : 1)];

	  else if (ind1[i1] < ind2[i2])
				// the member is in list 1 but not list 2, so it counts for either b or d
	    ++ x[f1[i1++] > 0 ? 1 : 3];

	  else 
				// the member is in list 2 but not list 1, so it counts for either c or d
	    ++ x[f2[i2++] > 0 ? 2 : 3];
	}
      // clean up rest of list 1
      while (i1 < f1.length)
				// the member is in list 1 but not list 2, so it counts for either b or d
	++ x[f1[i1++] > 0 ? 1 : 3];

      // clean up rest of list 2
      while (i2 < f2.length)
				// the member is in list 2 but not list 1, so it counts for either c or d
	++ x[f2[i2++] > 0 ? 2 : 3];

      a=x[0]; b=x[1]; c=x[2]; d=x[3];
    }

  public final void getCounts (Object o1, Object o2)
    {
      if (o1 instanceof float[] && o2 instanceof float[])
	getCounts ((float[]) o1, (float[]) o2);
      else if (o1 instanceof SparseVector && o2 instanceof SparseVector)
	getCounts ((SparseVector) o1, (SparseVector) o2);
      else
	a = b = c = d = 0;
    }

}

class distJaccard extends distFromBinarySim
{
  public final double distance (Object o1, Object o2)
    {
      getCounts (o1, o2);
      // return 1.0 - the Jaccard similarity:

      return (b + c) / ((double) (a + b + c));
    }
}

class distSorensen extends distFromBinarySim
{
  public final double distance (Object o1, Object o2)
    {
      getCounts (o1, o2);
      // return 1.0 - the Sorensen similarity:

      return (b + c) / ((double) (2 * a + b + c));
    }
}

class distSimpleMatching extends distFromBinarySim
{
  public final double distance (Object o1, Object o2)
    {
      getCounts (o1, o2);
      // return 1.0 - the Sorensen similarity:

      return (b + c) / ((double) (a + b + c + d));
    }
}

class distBaroniUrbani extends distFromBinarySim
{
  public final double distance (Object o1, Object o2)
    {
      getCounts (o1, o2);
      // return 1.0 - the Baroni-Urbani and Buser similarity:

      return (b + c) / (Math.sqrt (a * d) + a + b + c);
    }
}

class distEuclidean extends distMeasure
{
  public final double distance (Object o1, Object o2)
    {
      // return the Euclidean distance:

      if (o1 instanceof float[] && o2 instanceof float[])
	{
	  float[] v1 = (float[]) o1;
	  float[] v2 = (float[]) o2;

	  double dist = 0.0;
	  for (int i=0; i < v1.length; ++i)
	    {
	      double f = v1[i] - v2[i];
	      dist += f * f;
	    }
	  return Math.sqrt (dist);
	}
      return 0;
    }
}

class distMeanCensoredEuclidean extends distMeasure
{
  public final double distance (Object o1, Object o2)
    {
      // return the mean Euclidean distance, not counting
      // double zero pairs:

      if (o1 instanceof float[] && o2 instanceof float[])
	{
	  float[] v1 = (float[]) o1;
	  float[] v2 = (float[]) o2;

	  double dist = 0.0;
	  int n = 0;
	  for (int i=0; i < v1.length; ++i)
	    {
	      if (v1[i] != 0 || v2[i] != 0)
		{
		  ++n;
		  double f = v1[i] - v2[i];
		  dist += f * f;
		}
	    }
	  return (n > 0) ? (Math.sqrt (dist / n)) : 0;
	}
      return 0;
    }
}

class distBrayCurtis extends distMeasure
{
  public final double distance (Object o1, Object o2)
    {
      // return the Bray-Curtis distance:

      if (o1 instanceof float[] && o2 instanceof float[])
	{
	  float[] v1 = (float[]) o1;
	  float[] v2 = (float[]) o2;

	  double num = 0.0, denom = 0.0;
	  for (int i=0; i < v1.length; ++i)
	    {
	      if (v1[i] >= v2[i])			// inline the absolute value computation
		num += v1[i] - v2[i];
	      else
		num += v2[i] - v1[i];
				
	      denom += v1[i] + v2[i];
	    }
	  return num / denom;
	}
      return 0;
    }
}

class distCanberra extends distMeasure
{
  public final double distance (Object o1, Object o2)
    {
      // return the Canberra distance:

      if (o1 instanceof float[] && o2 instanceof float[])
	{
	  float[] v1 = (float[]) o1;
	  float[] v2 = (float[]) o2;

	  double can = 0.0;
	  int i;
	  for (i = 0; i < v1.length; ++i)
	    {
	      if (v1[i] > 0 || v2[i] > 0)
		{
		  double num;
		  if (v1[i] >= v2[i])			// inline the absolute value computation
		    num = v1[i] - v2[i];
		  else
		    num = v2[i] - v1[i];
					
		  can += num / (v1[i] + v2[i]);
		}
	    }
	  return can / i;
	}
      return 0;
    }
}

class OptionFrame extends Frame
{
  public OptionFrame (String s)
    {
      super (s);
    }

  public boolean handleEvent (Event evt)
    {
      if (evt.id == Event.WINDOW_DESTROY || evt.id == Event.WINDOW_ICONIFY)
	{
	  hide ();
	  return true;
	}
      else
	return super.handleEvent (evt);
    }
}


public class cluster extends Applet
  implements Runnable
{

  CalcRunner		crCalculate;
  Button			btnClearInputs;
  Button			btnShowOptions;
  Choice			chClusterMethod;
  Choice			chDistanceMeasure;
  Choice			chInputFormat;
  Choice			chNameLocation;
  Choice                       chStabilityAnalysis; 
  Choice                chSupportType;
  OptionFrame		ofOptionFrame;
  Checkbox		cbEmptyCells;
  Checkbox		cbHasDiagonal;
  Checkbox	        cbHasCharNames;
  Checkbox		cbOutputAsSimilarity;
  Checkbox		cbOutputDistances;
  Checkbox		cbOutputDistanceTable;
  Checkbox		cbOutputProgress;
  Checkbox		cbOutputTree;
  Checkbox		cbOutputPhylip;
  Checkbox              cbOutputGeneratedTrees;
  Label			labDistanceMeasure;
  Label			labClusterMethod;
  Label			labInputFormat;
  Label			labNameLocation;
  Label                 labStabilityAnalysis;
  Label                 labNumRuns;
  Label                 labSupportType;
  NumInputBox		nibSampleNames;
  NumInputBox		nibSampleData;
  NumOutputBox	nobOutput;
  TextField             tfNumRuns;
  String			output;
  Ranecu                rng;

  final static int CORNELL_CONDENSED = 0;
  final static int SQUARE_SIMILARITY = 1;
  final static int UPPER_TRIANGLE_SIMILARITY = 2;
  final static int LOWER_TRIANGLE_SIMILARITY = 3;
  final static int SQUARE_DISTANCE = 4;
  final static int UPPER_TRIANGLE_DISTANCE = 5;
  final static int LOWER_TRIANGLE_DISTANCE = 6;
  final static int SAMPLE_CHARACTER = 7;
  final static int CHARACTER_SAMPLE = 8;

  final static int SINGLE_LINKAGE = 0;
  final static int COMPLETE_LINKAGE = 1;
  final static int UNWEIGHTED_ARITHMETIC_AVERAGE = 2;
  final static int WEIGHTED_ARITHMETIC_AVERAGE = 3;
  final static int UNWEIGHTED_CENTROID = 4;
  final static int WEIGHTED_CENTROID = 5;
  final static int WARDS_MINIMUM_VARIANCE = 6;
  final static int NEIGHBOUR_JOINING = 7;

  final static int NAMES_IN_NAMES = 0;
  final static int NAMES_IN_DATA = 1;
  final static int NO_NAMES = 2;

  final static int STAB_NONE = 0;
  final static int STAB_JACKNIFE = 1;
  final static int STAB_BOOTSTRAP = 2;
  final static int STAB_PERMUTE = 3;
  final static int STAB_POISSON = 4;
  final static int STAB_MULTINOMIAL = 5;

  // options for how to compute the support of a node in a collection of trees
  // SUPP_JOIN is the traditional "support for a node":  N2 in T2 supports N1 in T1
  // if N2 has exactly the same set of descendent leaf nodes as N1 does
  // i.e. if leaves(n1) = leaves(N2)
  // SUPP_TRI is perhaps more appropriate when nodes are treated as dividing their
  // descendents into two groups.  N2 = (N21,N22) in T2 supports N1 = (N11,N12) in T1 if
  //     leaves(N21) = leaves(N11) and leaves(N22) = leaves(N12)
  // or
  //     leaves(N21) = leaves(N12) and leaves(N22) = leaves(N11)

  final static int SUPP_BI = 0;
  final static int SUPP_TRI = 1;

  public cluster()
    {
      // cluster Class Constructor
      //--------------------------------------------------------------------------


      setLayout(new GridBagLayout ());

      int r;		// variable for row of layout
      rng = null;

      Font f = new Font ("Times", Font.BOLD, 12);

      labInputFormat		= new Label ("Input data is: ", Label.RIGHT);
      labInputFormat.setFont (f);
      MyFuns.constrain (this,
			labInputFormat,
			0, r=0, 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chInputFormat		= new Choice ();
      chInputFormat.addItem ("Cornell condensed format");
      chInputFormat.addItem ("Square Similarity Matrix");
      chInputFormat.addItem ("Upper Triangular Similarity Matrix");
      chInputFormat.addItem ("Lower Triangular Similarity Matrix");
      chInputFormat.addItem ("Square Distance Matrix");
      chInputFormat.addItem ("Upper Triangular Distance Matrix");
      chInputFormat.addItem ("Lower Triangular Distance Matrix");
      chInputFormat.addItem ("Character Matrix: samples are columns");
      chInputFormat.addItem ("Character Matrix: samples are rows");
      MyFuns.constrain (this,
			chInputFormat,
			1, r, 4, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);
      chInputFormat.select (0);

      labNameLocation		= new Label ("Location of Sample Names:", Label.RIGHT);
      labNameLocation.setFont (f);
      MyFuns.constrain (this,
			labNameLocation,
			0, r += 1 , 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chNameLocation		= new Choice ();
      chNameLocation.addItem ("Names Window");
      chNameLocation.addItem ("Data Window");
      chNameLocation.addItem ("No sample names - use numbers");

      MyFuns.constrain (this,
			chNameLocation,
			1, r, 4, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chNameLocation.select (0);

      labDistanceMeasure		= new Label ("Distance/Similarity Measure:", Label.RIGHT);
      labDistanceMeasure.setFont (f);
      MyFuns.constrain (this,
			labDistanceMeasure,
			0, r += 1 , 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chDistanceMeasure		= new Choice ();
      for (int i=0; i < distMeasure.getNumMeasures (); ++i)
	chDistanceMeasure.addItem (distMeasure.getName (i));

      MyFuns.constrain (this,
			chDistanceMeasure,
			1, r, 4, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chDistanceMeasure.select (0);

      labClusterMethod		= new Label ("Clustering Method:", Label.RIGHT);
      labClusterMethod.setFont (f);
      MyFuns.constrain (this,
			labClusterMethod,
			0, r += 1 , 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chClusterMethod		= new Choice ();
      chClusterMethod.addItem ("Single Linkage");
      chClusterMethod.addItem ("Complete Linkage");
      chClusterMethod.addItem ("Unweighted Arithmetic Average");
      chClusterMethod.addItem ("Weighted Arithmetic Average");
      chClusterMethod.addItem ("Unweighted Centroid");
      chClusterMethod.addItem ("Weighted Centroid");
      chClusterMethod.addItem ("Ward's Minimum Variance");
      chClusterMethod.addItem ("Saitou and Nei Neighbour Joining");

      MyFuns.constrain (this,
			chClusterMethod,
			1, r, 4, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chClusterMethod.select (2);

      labStabilityAnalysis  = new Label ("Stability Analysis:", Label.RIGHT);
      labStabilityAnalysis.setFont (f);
      MyFuns.constrain (this,
			labStabilityAnalysis,
			0, r += 1 , 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chStabilityAnalysis		= new Choice ();
      chStabilityAnalysis.addItem ("None");
      chStabilityAnalysis.addItem ("Jacknife");
      chStabilityAnalysis.addItem ("Bootstrap");
      chStabilityAnalysis.addItem ("Randomize sample order");
      chStabilityAnalysis.addItem ("Resample - poisson");
      chStabilityAnalysis.addItem ("Resample - multinomial");

      MyFuns.constrain (this,
			chStabilityAnalysis,
			1, r, 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chStabilityAnalysis.select (1);

      labNumRuns		= new Label ("# Runs:", Label.RIGHT);
      labNumRuns.setFont (f);
      labNumRuns.enable (false);

      MyFuns.constrain (this,
			labNumRuns,
			2, r, 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      tfNumRuns		= new TextField ("100",8);

      MyFuns.constrain (this,
			tfNumRuns,
			3, r, 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      tfNumRuns.enable (false);

      labSupportType		= new Label ("Support Type:", Label.RIGHT);
      labSupportType.setFont (f);
      MyFuns.constrain (this,
			labSupportType,
			0, r += 1 , 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chSupportType		= new Choice ();
      chSupportType.addItem ("traditional:  group=all descendents of node");
      chSupportType.addItem ("symmetrical:  groups=left, right descendents");

      MyFuns.constrain (this,
			chSupportType,
			1, r, 1, 1,
			GridBagConstraints.HORIZONTAL,
			GridBagConstraints.WEST,
			1.0, 1.0, 2, 2, 2, 2
			);

      chSupportType.select (0);

      nibSampleNames		= new NumInputBox (this, 
						   "Sample Names",
						   "Enter sample names, separated by spaces, tabs, or newlines",
						   1,
						   5, 0, 1, 4,
						   GridBagConstraints.BOTH,
						   GridBagConstraints.NORTH
						   );


      btnShowOptions = new Button ("  More Options...  ");
      MyFuns.constrain (this,
			btnShowOptions,
			0, ++r, 1, 1,
			GridBagConstraints.NONE,
			GridBagConstraints.CENTER,
			1.0, 1.0, 2, 2, 2, 2
			);
	
      btnClearInputs = new Button ("  Clear Inputs  ");
      MyFuns.constrain (this,
			btnClearInputs,
			1, r, 1, 1,
			GridBagConstraints.NONE,
			GridBagConstraints.CENTER,
			1.0, 1.0, 2, 2, 2, 2
			);
	
      crCalculate = new CalcRunner (this, this, "  Calculate!  ", "  Cancel!  ");
      MyFuns.constrain (this,
			crCalculate,
			2, r, 2, 1,
			GridBagConstraints.NONE,
			GridBagConstraints.CENTER,
			1.0, 1.0, 2, 2, 2, 2
			);

      nibSampleData		= new NumInputBox (this, 
						   "Input Data",
						   "Enter data in the format you select",
						   2,
						   0, r += 5, 5, 1,
						   GridBagConstraints.BOTH,
						   GridBagConstraints.NORTH
						   );


      nobOutput		= new NumOutputBox (this,
					    "Output",
					    "Use copy/paste to move the results into other programs.",
					    2,
					    0, r += 5, 5, 1,
					    GridBagConstraints.BOTH,
					    GridBagConstraints.WEST,
					    20.0, 20.0, 1, 1, 1, 1
					    );


      /*
	Now create the extra options frame 
      */

      ofOptionFrame = new OptionFrame ("Other Clustering Options");
      ofOptionFrame.setLayout (new GridBagLayout ());

      r = 0;
      cbOutputGeneratedTrees = new Checkbox ("Stability analysis: show generated trees", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbOutputGeneratedTrees,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbHasCharNames = new Checkbox ("For matrix, row and column names are both present", null, true);
      MyFuns.constrain (ofOptionFrame,
			cbHasCharNames,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbEmptyCells = new Checkbox ("For input, items are separated by TABs, and zeros are omitted", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbEmptyCells,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbEmptyCells.enable (false);

      cbHasDiagonal = new Checkbox ("For triangular matrix input, 0 or 1 on diagonal are present", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbHasDiagonal,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbHasDiagonal.enable (true);


      cbOutputAsSimilarity = new Checkbox ("For output, show similarities instead of distances", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbOutputAsSimilarity,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbOutputDistances = new Checkbox ("Output the lower-triangular matrix of distances or similarities", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbOutputDistances,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);
      
      cbOutputDistanceTable = new Checkbox ("Output a table of distances or similarities", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbOutputDistanceTable,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbOutputProgress = new Checkbox ("Output the progress of the clustering algorithm", null, true);
      MyFuns.constrain (ofOptionFrame,
			cbOutputProgress,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbOutputTree = new Checkbox ("Output the cluster tree", null, true);
      MyFuns.constrain (ofOptionFrame,
			cbOutputTree,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      cbOutputPhylip = new Checkbox ("Output a Phylip/DRAWTREE-readable version of the tree", null, false);
      MyFuns.constrain (ofOptionFrame,
			cbOutputPhylip,
			0, ++r, 4, 1,
			GridBagConstraints.BOTH,
			GridBagConstraints.EAST,
			1.0, 1.0, 2, 2, 2, 2
			);

      ofOptionFrame.setResizable (false);
      ofOptionFrame.pack ();
      ofOptionFrame.move (100, 100);

    }
  
  public String getAppletInfo()
    {
      return "Name: cluster\r\n" +
	"Author: John Brzustowski\r\n" +
	"";
    }
  

public void destroy()
{
  ofOptionFrame.dispose ();
}

public void paint(Graphics g)
{
}

public void start()
{
}
	
public void stop()
{
}

public boolean handleEvent (Event evt)
{
  if (evt.id == Event.ACTION_EVENT)
    {
      if (evt.target instanceof Button )
	{
	  if (evt.target == btnClearInputs )
	    {
	      nibSampleData.setText ("");
	      nibSampleNames.setText ("");
	      return true;
	    }
	  if (evt.target == btnShowOptions)
	    {
	      ofOptionFrame.show ();
	      ofOptionFrame.toFront ();
	      return true;
	    }
	} 
      else if (evt.target == chInputFormat)
	{
	  super.handleEvent (evt);		// let user select input format,
	  // then see if distance measure is valid
	  int i=chInputFormat.getSelectedIndex ();
	  switch (i)
	    {
	    case SAMPLE_CHARACTER:
	    case CHARACTER_SAMPLE:
	      cbEmptyCells.enable (true);
	      chDistanceMeasure.enable (true);
	      cbHasDiagonal.enable (false);
	      chStabilityAnalysis.enable (true);
	      setStabilityControls (); // then enable/disable relevant controls
	      break;
	    case CORNELL_CONDENSED:
	      cbEmptyCells.enable (false);
	      chDistanceMeasure.enable (true);
	      cbHasDiagonal.enable (false);
	      chStabilityAnalysis.enable (false);
	      tfNumRuns.enable (false);
	      labNumRuns.enable (false);
	      break;
	    default:
	      chDistanceMeasure.enable (false);
	      cbEmptyCells.enable (false);
	      cbHasDiagonal.enable (i != SQUARE_DISTANCE && i != SQUARE_SIMILARITY);
	      chStabilityAnalysis.enable (false);
	      tfNumRuns.enable (false);
	      labNumRuns.enable (false);
	      break;
	    }
	  return true;
	}
      else if (evt.target == chStabilityAnalysis)
	{
	  super.handleEvent (evt); // let user select input format,
	  setStabilityControls (); // then enable/disable relevant controls
	  return true;
	}
    }
  return false;
}

public void setStabilityControls ()
{
  int i=chStabilityAnalysis.getSelectedIndex ();
  switch (i)
    {
    case STAB_NONE:
      labSupportType.enable (false);
      chSupportType.enable (false);
      tfNumRuns.enable (false);
      labNumRuns.enable (false);
      break;
    case STAB_JACKNIFE:
      labSupportType.enable (true);
      chSupportType.enable (true);
      tfNumRuns.enable (false);
      labNumRuns.enable (false);
      break;
    default:
      tfNumRuns.enable (true);
      labNumRuns.enable (true);
      labSupportType.enable (true);
      chSupportType.enable (true);
      break;
    }
}

public void computeSampleDistances (ProgressBar pb, float values[][], double distances[][], int numSamples, int numChars, int distanceMeasure)
{
  distMeasure dm = distMeasure.getDistMeasure (distanceMeasure);
  
  pb.SetTotalCount (numSamples * numChars * numSamples / 2);
  
  for (int i=0; i < numSamples; ++i)
    {
				// compute distance from this vector to previous ones
      for (int j=0; j < i; ++j)
	distances[i][j] = dm.distance (values[i], values[j]);
      
      pb.ShowProgress (i);
    }		
}

public void run () throws RuntimeException
{
  output = "";
  nobOutput.setText (output);

  String EOL = nobOutput.getEOL ();
  String delim = nobOutput.getDelim ();

  // Note:  the algorithm used for SAHN clustering is a translation into Java
  // of the Fortran code by Fionn Murtagh (fmurtagh@eso.org) available
  // in the package http://www.stat.cmu.edu/S/multiv
		
  // The source carries this notice:
  /*
    Notes: These routines have been used on a Sun Unix platform with S-Plus v. 3.1.
    This software can be freely used and freely distributed for non-commercial
    purposes.                                      Fionn Murtagh (fmurtagh@eso.org)
  */


  int numSamples;			// the number of individuals, samples, or populations
  int numChars = -1;		// number of characters per sample, only for the Sample<->Character Matrices

  String sampleData = nibSampleData.getText ();
  int clusterMethod = chClusterMethod.getSelectedIndex ();
  int nameLocation = chNameLocation.getSelectedIndex ();
  int distanceMeasure = chDistanceMeasure.getSelectedIndex ();
  int supportType = chSupportType.getSelectedIndex ();

  boolean asSimilarity = cbOutputAsSimilarity.getState ();
  boolean emptyCells = cbEmptyCells.getState ();
  boolean namesInData = nameLocation == NAMES_IN_DATA;
  boolean hasDiagonal = cbHasDiagonal.getState ();
  boolean hasCharNames = cbHasCharNames.getState ();
  boolean outputGeneratedTrees = cbOutputGeneratedTrees.getState ();
  int inputFormat = chInputFormat.getSelectedIndex ();
  ProgressBar pb = new ProgressBar (this, "Reading input");

  ELTree clustTree = null;

  int numInputTokens = nibSampleData.countValues ();
  int numTokensNeeded = -1;		// how many items are expected in the input window (-1 means can't tell)
  // we use this to tell if the user hasn't supplied the correct number of tokens

  int stabilityAnalysis = STAB_NONE; // by default, no tree stability analysis
  float [][] values = null;  // matrix values, to keep for stability analysis

  try
    {
      // figure out how many samples there are
			
      switch (inputFormat)
	{
	case CORNELL_CONDENSED:
	  numSamples = countCornellSamples (sampleData);
	  break;
	case SQUARE_SIMILARITY:
	case SQUARE_DISTANCE:
				// even if names appear as the first token in each line,
				// the following yields the correct value

	  numSamples = (int) Math.sqrt (numInputTokens);
	  numTokensNeeded = numSamples * (numSamples + (namesInData ? 1 : 0));
	  break;
	case SAMPLE_CHARACTER:
				// the first token in the input is the number of characters per sample
	  try
	    {
	      numChars =  Integer.parseInt ((new StringTokenizer (sampleData)).nextToken ());
	    }
	  catch (Exception e)
	    {
	      throw new Exception ("For a Samples-as-Columns matrix,\nthe first item in the input window must be\nthe number of characters.\n");
	    }
				// the number of samples depends on whether names are included in the data window, 
				// and on whether empty cells are marked by tabs
	  if (emptyCells)
	    {
	      int numTabs = MyFuns.CountCharsInString (sampleData, '\t');
	      numSamples = 1 + (numTabs - (hasCharNames ? numChars : 0) - ((hasCharNames && namesInData) ? 1 : 0)) / (numChars + (namesInData ? 1 : 0));
	      numTokensNeeded = 0;		// can't easily tell how many tokens needed
	    }
	  else
	    {
	      numSamples = (numInputTokens - 1 - (hasCharNames ? numChars : 0) ) / (numChars + (namesInData ? 1 : 0));
	      numTokensNeeded = 1 + numSamples * (numChars + (namesInData ? 1 : 0)) + (hasCharNames ? numChars : 0);
	    }
	  break;
	case CHARACTER_SAMPLE:
				// the first token in the input is the number of characters per sample
	  try
	    {
	      numChars =  Integer.parseInt ((new StringTokenizer (sampleData)).nextToken ());
	    }
	  catch (Exception e)
	    {
	      throw new Exception ("For a Samples-as-Rows matrix,\nthe first item in the input window must be\nthe number of characters.\n");
	    }
				// the number of samples depends on whether names are included in the data window, 
				// and on whether empty cells are marked by tabs
	  if (emptyCells)
	    {
	      int numTabs = MyFuns.CountCharsInString (sampleData, '\t');
	      numSamples = (numTabs-(hasCharNames ? numChars - 1 : 0) - ((hasCharNames && namesInData) ? 1 : 0)) / (numChars - 1 + (namesInData ? 1 : 0));
	      numTokensNeeded = 0;		// can't easily tell how many tokens needed
	    }
	  else
	    {
	      // take into account any names present
	      numSamples = (numInputTokens - 1 - (hasCharNames ? numChars : 0) ) / (numChars + (namesInData ? 1 : 0));
	      numTokensNeeded = 1 + numSamples * (numChars + (namesInData ? 1 : 0)) + (hasCharNames ? numChars : 0);
	    }
	  break;
	default:
				// the number of tokens is either n * (n+1) / 2
				// or n * (n-1) / 2, depending on whether or not names appear
				// as the first token in each line

	  numSamples = (int) Math.floor (Math.sqrt (2 * numInputTokens)) + 1;
	  if (namesInData)
	    --numSamples;
	  if (hasDiagonal)
	    --numSamples;
	  numTokensNeeded = numSamples * (numSamples-1) / 2 + (namesInData ? numSamples : 0) + (hasDiagonal ? numSamples : 0);
	}

      // crap out if there are too few samples
      if (numSamples < 2)		
	throw new Exception ("Input contains fewer than 2 samples.");

      // crap out if number of items in the input window is incorrect
      if (numTokensNeeded > 0 && numTokensNeeded != numInputTokens)
	throw new Exception ("I counted " + numSamples + " samples" + ((numChars >= 0) ? " and " + numChars + " characters." : ".") + "\nThis requires " + numTokensNeeded + " items in the input window,\nbut it contains " + numInputTokens + ".\nThere is probably data missing from your input.");
			
      // allocate space for intermediates and results

      double distances[][] = new double [numSamples][];		// a lower triangular (diagonal included) array of distances
      for (int i = 0; i < numSamples; ++i)
	distances[i] = new double [i + 1];
				
      String sampleNames[] = new String [numSamples];			// an array of sample names

      int clustNode1[] = new int [numSamples-1];				// indices of first cluster joined at step i
      int clustNode2[] = new int [numSamples-1];				// indices of second cluster joined at step i
			
      double clustDist[]  = null;				// distance of nodes joined at step i
      double mergeDist1[] = null;				// distance of node 1 to merger, for Neighbour joining
      double mergeDist2[] = null;				// distance of node 2 to merger, for Neighbour joining

      int samplePermutation[] = null;;

      // create the distance matrix, according to the selected input format

      switch (inputFormat)
	{
	case CORNELL_CONDENSED:
	  CornellToDistance (pb, sampleData, distances, sampleNames, namesInData, distanceMeasure);
	  break;
	case SQUARE_SIMILARITY:
	case SQUARE_DISTANCE:
	  MatrixToLowerTriangle (pb, sampleData, numInputTokens, distances, sampleNames, namesInData, numSamples, -1, false, false);
	  break;
	case UPPER_TRIANGLE_SIMILARITY:
	case UPPER_TRIANGLE_DISTANCE:
	  MatrixToLowerTriangle (pb, sampleData, numInputTokens, distances, sampleNames, namesInData, 0, 0, true, hasDiagonal);
	  break;
	case LOWER_TRIANGLE_SIMILARITY:
	case LOWER_TRIANGLE_DISTANCE:
	  MatrixToLowerTriangle (pb, sampleData, numInputTokens, distances, sampleNames, namesInData, hasDiagonal ? 1 : 0, 0, false, hasDiagonal);
	  break;
	case SAMPLE_CHARACTER:
	case CHARACTER_SAMPLE:
	  stabilityAnalysis = chStabilityAnalysis.getSelectedIndex ();

	  values = CharMatrixFromText (pb, sampleData, numSamples, numChars, sampleNames, namesInData, hasCharNames, inputFormat == SAMPLE_CHARACTER, emptyCells);

	  computeSampleDistances (pb, values, distances, numSamples, numChars, distanceMeasure);
	  break;
	}

      // convert similarities to distances

      switch (inputFormat)
	{
	case SQUARE_SIMILARITY:
	case UPPER_TRIANGLE_SIMILARITY:
	case LOWER_TRIANGLE_SIMILARITY:
	  for (int i=0; i < numSamples; ++i)
	    for (int j=0; j < distances[i].length - 1; ++j)
	      distances[i][j] = 1.0 - distances[i][j];
	  break;
	}

      // grab names from Names window, if any are there

      if ((!namesInData) && inputFormat != CORNELL_CONDENSED)
	{
	  StringTokenizer tok = new StringTokenizer ((nameLocation == NO_NAMES) ? "" : nibSampleNames.getText (), "\r\n \t");
	  for (int i=0; i < numSamples; ++i)
	    if (tok.hasMoreTokens ())
	      sampleNames[i] = tok.nextToken ();
	    else
	      sampleNames[i] = String.valueOf (i+1);
	}

      pb = new ProgressBar (this, "Cluster");

      output = "Cluster V0.1" + EOL + EOL + "Using " + chClusterMethod.getItem (clusterMethod) + EOL ;
			
      if (chDistanceMeasure.isEnabled ())
	output += "The Distance/Similarity measure is " + chDistanceMeasure.getSelectedItem () + "." + EOL;

      output += "There are " + numSamples + " samples." + EOL;

      if (cbOutputDistances.getState ())	// must do this before clustering, since distance matrix gets clobbered
	{
	  output += (asSimilarity ? "Similarity Matrix:" : "Distance matrix:") + EOL;

				// put out a line of column headers
	  for (int i=0; i < numSamples - 1; ++i)
	    output += delim + sampleNames[i];
	  output += EOL;

	  for (int i=1; i < numSamples; ++i)
	    {
	      output += sampleNames[i] + delim;
	      if (asSimilarity)
		for (int j=0; j < i; ++j)
		  output += String.valueOf (1 - distances[i][j]) + delim;
	      else
		for (int j=0; j < i; ++j)
		  output += String.valueOf (distances[i][j]) + delim;
	      output += EOL;
	    }
	  output += EOL;
	}

      if (cbOutputDistanceTable.getState ())	// must do this before clustering, since distance matrix gets clobbered
	{
	  output += "From" + delim + "To" + delim + (asSimilarity ? "Similarity" : "Distance  ") + EOL;
	  output += "========================" + EOL;
			
	  for (int i=0; i < numSamples; ++i)
	    {
	      for (int j=0; j < i; ++j)
		output += sampleNames[i] + delim + sampleNames[j] + delim + String.valueOf (asSimilarity ? 1 - distances[i][j] : distances[i][j]) + EOL;
	    }
	  output += EOL;
	}

      if (cbOutputProgress.getState () || cbOutputTree.getState () || cbOutputPhylip.getState ())
	{
	
	  if (clusterMethod != NEIGHBOUR_JOINING)
	    {
	      clustDist = new double [numSamples-1];			// distance of nodes joined at step i
	      SAHNClustering (pb, clusterMethod, clustNode1, clustNode2, clustDist, distances);
	      if (cbOutputProgress.getState ())
		{
		  output += "Clustering procedure:" + EOL;
		  output += "Step       A      B  " + (asSimilarity ? "Similarity" : "  Distance") + EOL;
		  output += "===============================" + EOL;
		  for (int i=0; i < numSamples-1; ++i)
		    {
		      output += (i+1) + delim 
			+ (clustNode1[i] >= 0 ? sampleNames[clustNode1[i]] : String.valueOf (- clustNode1[i]) + '\'') + delim 
			+ (clustNode2[i] >= 0 ? sampleNames[clustNode2[i]] : String.valueOf (- clustNode2[i]) + '\'') + delim
			+ (asSimilarity ? 1 - clustDist[i] : clustDist[i]) + EOL;
		    }
		}
	    }
	  else
	    {
	      mergeDist1 = new double [numSamples-1];			// distances of nodes i and j
	      mergeDist2 = new double [numSamples-1];			// to the new node formed by fusing them
	      NeighbourJoining (pb, clustNode1, clustNode2, mergeDist1, mergeDist2, distances);
	      if (cbOutputProgress.getState ())
		{
		  output += "Neighbour-Joining procedure:" + EOL;
		  output += "Step       A    Length    B   Length" + EOL;
		  output += "======================================" + EOL;
		  for (int i=0; i < numSamples-1; ++i)
		    {
		      output += (i+1) + delim 
			+ (clustNode1[i] >= 0 ? sampleNames[clustNode1[i]] : String.valueOf (- clustNode1[i]) + '\'') + delim 
			+ mergeDist1[i] + delim 
			+ (clustNode2[i] >= 0 ? sampleNames[clustNode2[i]] : String.valueOf (- clustNode2[i]) + '\'') + delim
			+ mergeDist2[i] + EOL;
		    }
		}
	    }


	  if (cbOutputTree.getState () || cbOutputPhylip.getState ())
	    {
	      ELTree.setInodeSuffix ("'");
	      if (clusterMethod != NEIGHBOUR_JOINING)
		clustTree = ELTree.fromClusterOutput (clustNode1, clustNode2, clustDist, sampleNames, null);
	      else
		clustTree = ELTree.fromNeighbourOutput (clustNode1, clustNode2, mergeDist1, mergeDist2, sampleNames);

	      clustTree.setSeparatorChars (delim, EOL);
	      clustTree.sortByLeaves ();

	      if (cbOutputTree.getState ())
		{
		  output += EOL + "Tree Topology (edge lengths ignored)" + EOL + EOL +  clustTree.toTextPicture ();
		  output += EOL + "From    To        Length";
		  output += EOL + "========================";
		  output += EOL + clustTree.lengthTable () + EOL;
		}
	      if (cbOutputPhylip.getState ())
		output += "Tree in Phylip/DRAWTREE format: " + EOL +  clustTree.toString () + ";" + EOL;
	    }
	}
      nobOutput.setText (output);

      // a shovelful of straight line code to do stability analysis on the observed tree
      //
      // repeat numRuns times:
      //    - generate a new matrix of character values for each sample (either jacknifed or bootstrapped)
      //    - calculate intersample distances for the new matrix
      //    - build a tree by clustering according to the new distances
      //    - compare the resulting tree to the "observed" tree (i.e. the one built from the original data)
      //    - for each node in the observed tree, check whether the bipartition it induces in the leaf nodes
      //      is also present in the new tree
      // - output the original tree with frequencies at each node, indicating how often that node's induced partition
      //   was found in the generated trees

      if (stabilityAnalysis != STAB_NONE)
	{
	  output += EOL + "About to do stability analysis..." + EOL;
	  nobOutput.setText (output);

	  int numRuns;
	  
	  if (stabilityAnalysis == STAB_JACKNIFE)
	    numRuns = numChars;
	  else
	    {
	      rng = new Ranecu(); // set up a random number generator
	      try
		{
		  numRuns =  Integer.parseInt (tfNumRuns.getText ());
		  if (numRuns < 0)
		    numRuns = 0;
		}
	      catch (Exception e)
		{
		  output +=  "Error: for stability analysis, the number of runs must be an integer.\n";
		  numRuns = 0;
		}
	    }
	      
	  float newValues[][] = null;	// array for newly generated character/sample values
	  ELTree newTree;	// cluster tree for new data
	  int numInodes = numSamples - 1 - ((clusterMethod == NEIGHBOUR_JOINING) ? 1 : 0);
	  int supportCounts[] = new int [numInodes]; // array to count how many times each partition in the 
				// original tree appeared in a generated tree

	  // we use the original distances[][] array - the original values aren't needed any more

	  // set up some values required for support calculation - to avoid having to do this each time

	  clustTree.sortByLeaves ();
	  oLeaves = new int [numSamples];
	  clustTree.getLeafNodes (oLeaves, 0);
	  
	  oLeafPos = new int [numSamples];
	  for (int i=0; i < numSamples; ++i)
	    oLeafPos[oLeaves[i]] = i;

	  oFirst = new int [numInodes];
	  oLast = new int [numInodes];
	  clustTree.getTreeExtents (oFirst, oLast, 0);
	  
	  oHashValues = new int [numInodes];
	  clustTree.hashByLeaves (oHashValues, clusterMethod == NEIGHBOUR_JOINING, supportType == SUPP_BI);
	  
	  oSortedInode = new int [numInodes];
	  radixSortHashValues (oHashValues, oSortedInode);
	  
	  oInodes = new ELTree [numInodes];
	  clustTree.getInodes (oInodes);

	  // generate sums of characters for each sample

	  charSums = new double [numSamples];
	  for (int i=0; i < numSamples; ++i)
	    for (int j=0; j < numChars; ++j)
	      charSums[i] += values[i][j];

	  // here, we arbitrarly decide to truncate poisson distributions at the value 1000
	  // this should be user specifiable
	  // or the user should be given the choice to truncate a tail of the distribution
	  // and we would calculate the integer cutoff for the largest of the character values

	  int cutoff = 1000;

	  for (int r = 0; r < numRuns; ++ r )
	    {

	      pb.SetMessage ("Stability run #" + (1 + r));
	      // generate a new matrix of values for this run
	      switch (stabilityAnalysis)
		{
		case STAB_PERMUTE:
		  samplePermutation = new int[numSamples];
		  newValues= resamplePermute (values, rng, samplePermutation);
		  break;
 		case STAB_JACKNIFE:
		  newValues = resampleJacknife (values, r);
		  break;
 		case STAB_BOOTSTRAP:
		  newValues = resampleBootstrap (values, rng);
		  break;
		case STAB_MULTINOMIAL:
		  newValues = resampleMultinomial (values, rng);
		  break;
		case STAB_POISSON:
		  newValues = resamplePoisson (values, rng, cutoff);
		  break;
		}
	      
	      // dump the resampled matrix

	      /*
	      System.out.println("Values for run # " + (1+r));
	      for (int s=0; s < numSamples; ++s)
		{
		  for (int c=0; c < numChars; ++c)
		    System.out.print (newValues[s][c] + " ");
		  System.out.println ("");
		  
		}
	      */

	      // compute intersample distances

	      computeSampleDistances (pb, newValues, distances, numSamples, numChars, distanceMeasure);

	      // cluster the samples and generate the resulting edge-labelled tree
	      
	      if (clusterMethod != NEIGHBOUR_JOINING)
		{
		  SAHNClustering (pb, clusterMethod, clustNode1, clustNode2, clustDist, distances);
		  newTree = ELTree.fromClusterOutput (clustNode1, clustNode2, clustDist, sampleNames, samplePermutation);
		}
	      else
		{
		  NeighbourJoining (pb, clustNode1, clustNode2, mergeDist1, mergeDist2, distances);
		  newTree = ELTree.fromNeighbourOutput (clustNode1, clustNode2, mergeDist1, mergeDist2, sampleNames);
		}
	      if (outputGeneratedTrees)
		output += EOL + "Generated Tree #" + (r+1) + " Topology:" + EOL + EOL +  newTree.toTextPicture ();

	      // for each partition in the original tree, see if it exists in the new tree

	      if (clusterMethod != NEIGHBOUR_JOINING)
		{
		  seekSupport (clustTree, newTree, supportCounts, supportType);
		}
	      else
		{
		  seekSupportUnrooted (clustTree, newTree, supportCounts, supportType);
		}

	    }

	  // output a node stability table
	  // i.e. for each interior node in the observed tree, print the number of times
	  // its induced partition was seen in the generated trees

	  if (numRuns > 0)
	    {

	      if (clusterMethod == NEIGHBOUR_JOINING && supportType == SUPP_BI)
		--numInodes;

	      output += EOL + "Results of stability analysis:" + EOL;
	      output += "Runs: " + numRuns + ".  New data generated by " + chStabilityAnalysis.getSelectedItem () + EOL;
	      output += "Node" + delim + "Count" + EOL;
	      for (int i=0; i < numInodes; ++i)
		output += (i+1) + "'" + delim + supportCounts[i] + EOL;

	      if (clusterMethod != NEIGHBOUR_JOINING)
		{
		  output += EOL + "For each (interior) node in the original tree, the table" + EOL
		    + "says how many of the generated trees had an equivalent interior node." + EOL
		    + "Two interior nodes (in different trees) are equivalent if they" + EOL;
		  if (supportType == SUPP_BI)
		    output += "have the same samples as descendents." + EOL;
		  else
		    output += "have the same samples AND split those samples into the same two groups." + EOL;
		}
	      else
		{
		  if (supportType == SUPP_BI)
		    output += EOL + "For each interior edge in the original tree, the table" + EOL
		      + "says how many of the generated trees had an equivalent edge." + EOL
		      + "Two edges are equivalent if removing them splits each tree" + EOL
		      + "into the same two groups of samples.  The edge numbered 'i'" + EOL
		      + "is the edge leading from interior node 'i' toward the pseudo-root" + EOL
		      + "of the original tree.  Interior edges are those not adjacent to samples" + EOL
		      + "so, for example, an unrooted tree of four samples has one interior edge.";
		  else
		    output += EOL + "For each (interior) node in the original tree, the table" + EOL
		      + "says how many of the generated trees had an equivalent interior node." + EOL
		      + "Two interior nodes (in different trees) are equivalent if they" + EOL
		      + "split the samples into the same three groups.";
		}
	    }
	  nobOutput.setText (output);
	}
      showStatus ("Done.");
    }

  catch (Exception e)
    {
      nobOutput.setText (nobOutput.getText() + "\n" + e.toString() + e.getMessage () );
      return;
    }
}

	
  public void MatrixToLowerTriangle (ProgressBar pb, String s, int numTokens, double mat[][], String names[], boolean getNames, int skip, int deltaSkip, boolean transpose, boolean hasDiagonal) throws Exception
{
  // extract the appropriate lower triangle and possibly a list of names
  // from one of our input formats
	 
  StringTokenizer tok = new StringTokenizer (s, "\r\n\t ");
  if (numTokens == 0)
    numTokens = tok.countTokens ();

  pb.SetTotalCount (numTokens);

  try
    {
      for (int i = 0; i < mat.length; ++i)	// loop over data for each row
	{
	  int jStart, jEnd;
			
	  if (getNames)		// get the name of this row, if necessary
	    names[i] = tok.nextToken ();
				
			
	  if (i >= mat.length)
	    break;

				
	  if (transpose)
	    {
	      jStart = i + 1;
	      jEnd = mat.length;
	      if (hasDiagonal)
		tok.nextToken ();
	    }
	  else
	    {
	      jStart = 0;
	      jEnd = i;
	    }

	  for (int j = jStart; j < jEnd; ++j)
	    {
	      double d = Double.valueOf (tok.nextToken ()).doubleValue ();
	      if (transpose)
		mat[j][i] = d;
	      else
		mat[i][j] = d;
	    }

	  for (int k=0; k < skip; ++k)		// skip over tokens (for square matrices)
	    tok.nextToken ();
	  skip += deltaSkip;
	  pb.ShowProgress (skip + jEnd-jStart + (getNames ? 1 : 0));

	}
    }
  catch (Exception e)
    {
      if (e instanceof NumberFormatException)
	throw new Exception ("While reading input, expected a number but got: " + e.getMessage ());

      if (e instanceof NoSuchElementException)
	throw new Exception ("Not enough numbers in input.");

      throw new Exception ("While reading input, got unexpected error: " + e.getMessage ());
    }
}

		
		
public int countCornellSamples (String s)
{
  // figure out how many samples are in a Cornell condensed format string

  StringTokenizer tok = new StringTokenizer (s, "\r\n");
  int count = 0;
  String prevTag = "";
  for (int i = tok.countTokens (); i > 0; --i)
    {
      String t = (new StringTokenizer (tok.nextToken (), " \t")).nextToken ();
      if (! prevTag.equals (t))
	{
	  ++count;
	  prevTag = t;
	}
    }
  return count;
}


public void CornellToDistance (ProgressBar pb, String s, double distances[][], String names [], boolean getNames, int distanceMeasure) throws Exception
{
  // takes a Cornell condensed format string and returns a lower-triangular distance
  // matrix
  // If getNames == true, fills names[] with names

  distMeasure dm = distMeasure.getDistMeasure (distanceMeasure);
  pb.SetTotalCount (distances.length * (distances.length-1) / 2);
  SparseVector svec[] = new SparseVector [distances.length];
  StringTokenizer tok = new StringTokenizer (s, "\r\n");
  int count = 0;
  String prevTag = "";
  String line = "";
  String t = null;
  StringTokenizer tok2 = null;
  for (int i = tok.countTokens (); i >= 0; --i)
    {

      if (i > 0)
	{
	  tok2 = new StringTokenizer (tok.nextToken (), " \t");
	  t = tok2.nextToken ();
	}
      if (i > 0 && prevTag.equals (t))	// keep adding new line to buffer
	{
	  line += tok2.nextToken ("\0");
	}
      else 
	{
	  if (line.length() > 0)
	    {
	      svec[count] = new SparseVector (line);
	      if (getNames)
		names[count] = prevTag;
					
	      for (int k=0; k < count; ++k)
		distances[count][k] = dm.distance(svec[count], svec[k]);

	      pb.ShowProgress (count);

	      ++count;
	    }
	  prevTag = t;
	  if (i > 0)
	    line = tok2.nextToken ("\0");
	}
    }
}


public void NeighbourJoining (ProgressBar pb, int ia[], int ib[], double length1[], double length2[], double dist[][]) throws Exception
{
  // Saitou and Nei (1987) neighbour joining method
  // The algorithm is O(n^3), at least in the revision
  // presented by Studier and Keppler (1988)
		
  // (NEIGHBOR in the Phylip package is actually implemented
  // in a way that requires time o(N^4))
		
  /*
    Recalculation of r[] is optimized by using the following
    shortcuts, which come from the definitions
    of r[] and the new node distances.

    Suppose nodes i and j have just merged to form node u, and let
    k be any other node.  There are N nodes, counting
    i and j but not u.

    Then:

    r[u] = (r[i] + r[j] - N * d(i,j)) / 2

    and

    r'[k] = r[k] - d(i,k) - d(j,k) + d(k,u)
			
    where r'[] is the set of new divergences for nodes other than
    i, j, and u

    These formulas don't improve the order, because we
    still search over all j < i < N for the minimum
    value of Mij, but the constant factor multiplying N^3 is reduced.
  */

  // The algorithm can be made slightly more elegant
  // by running through N-2 iterations, rather than
  // N-3, and simplifying the endgame,
  // but I wanted the output tree to be exactly
  // the same as that from Phylip's NEIGHBOR for easy comparison.

  int numItems = dist.length;
		
  double r[] = new double[numItems];	// r[i] is sum of distances from node i to other nodes
  int nodeNum[] = new int[numItems];	// number of this node, negative means interior, zero means merged
	
  // compute divergence, r, the sum of distances to all other nodes

  for (int i=0; i < numItems; ++i)
    {
      nodeNum[i] = i+1;			// initial node numbers
      for (int j=0; j < i; ++j)
	{	
	  r[i] += dist[i][j];		// add the distance to r for
	  r[j] += dist[i][j];		// each node involved
	}
    }

  int N = numItems;
  int iMin = -1;
  int jMin = -1;
  int i;
  int j;
  int u = 0;

  pb.SetTotalCount (N * N / 2 - 5);

  while (N > 2)		// while there are at least three nodes left
    {
      // find the smallest value of 
      // Mij := (N-2) * dist[i][j] - (r[i]+ r[j])

      double MMin = Double.MAX_VALUE;

    iloop:		for (i=1; i < numItems; ++i)
      {
	if (nodeNum[i] == 0)			// skip if node has been merged
	  continue iloop;

      jloop:			for (j=0; j < i; ++j)
	{	
	  if (nodeNum[j] == 0)		// skip if node has been merged
	    continue jloop;

	  double M = (N-2) * dist[i][j] - (r[i]+r[j]);
	  if (M < MMin)
	    {
	      MMin = M;
	      iMin = i;
	      jMin = j;
	    }
	}
      }
      // note which nodes are being merged
      // (this ordering yields the same (planar) tree as Phylip's NEIGHBOR program)

      ia [numItems - N] = nodeNum [jMin] - (nodeNum[jMin] > 0 ? 1 : 0);	// reduce positive node numbers back to zero origin	
      ib [numItems - N] = nodeNum [iMin] - (nodeNum[iMin] > 0 ? 1 : 0);

      // replace node j with new node u, the merge of i and j
		
      u = jMin;		// for convenience, u is the index for the merged node

      nodeNum [u] = (-1) - (numItems - N);			// assign negative interior node number
      nodeNum [iMin] = 0;								// mark as merged

      // record distances from merged nodes to their merger
      // (making sure to match the order of nodes from above)

      double dij = dist[iMin][jMin];
      length2 [numItems - N] = (dij + (r[iMin] - r[jMin])/(N-2)) / 2;
      length1 [numItems - N] = dij - length2[numItems - N];

      /*
	I thought perhaps NEIGHBOR did this moving negative lengths
	to the adjacent node.  Doesn't seem like it does, so I've
	left this out.

	if (length1 [numItems - N] < 0)
	{
	length1 [numItems - N] = 0;
	length2 [numItems - N] = dij;
	}
	else if (length2 [numItems - N] < 0)
	{
	length2 [numItems - N] = 0;
	length1 [numItems - N] = dij;
	}
      */
  
      // compute divergence of new node

      r[u] = (r[iMin] + r[jMin] - dij * N) / 2;
		
      // compute distances to new nodes, 
      //		d(u,k) = (d(iMin,k) + d(jMin,k) - d(iMin,jMin)) / 2
		
      // and recompute divergences of other nodes
      //		r'(k) = r(k) + d(u,k) - d(iMin,k) - d(jMin,k)


    recalc:
      for (int k=0; k < numItems; ++k)
	{
	  if (nodeNum[k]==0 || k == u)	// don't recompute for merged nodes or new node
	    continue recalc;

	  // compute the distance d(u,k):
			
	  double duk;
	  double dik;
	  double djk;
			
				// get d(iMin, k)

	  if (iMin > k)						
	    dik = dist[iMin][k];
	  else
	    dik = dist[k][iMin];

				//  get d(jMin, k)

	  if (jMin > k)
	    djk = dist[jMin][k];
	  else
	    djk = dist[k][jMin];

				// subtract dij and divide by 2 to get d(u,k)

	  duk = (dik + djk - dij) / 2;

	  r[k] += duk - dik - djk;

				// we no longer need the old value of d(jMin,k)
				// so overwrite it with the new value d(u,k)

	  if (u > k)
	    dist[u][k] = duk;
	  else
	    dist[k][u] = duk;
	}

      pb.ShowProgress (N);
				
      --N;		// reduce number of nodes left
    }			// keep merging until only two nodes are left

  /* find the remaining node */

  for (i=0; i < numItems; ++i)
    if (i != u && ! (nodeNum [i] == 0))
      break;
	
  /* connect the remaining node to the 'root' */

  ia [numItems - N] = nodeNum [u] - (nodeNum[u] > 0 ? 1 : 0);		
  ib [numItems - N] = nodeNum [i] - (nodeNum[i] > 0 ? 1 : 0);
		
  if (i > u)
    length1[numItems - N] = length2[numItems - N] = dist[i][u];
  else
    length1[numItems - N] = length2[numItems - N] = dist[u][i];

  return;
}

public void SAHNClustering (ProgressBar pb, int iopt, int ia[], int ib[], double crit[], double diss[][]) throws Exception
{
  /*
    a translation of Fionn Murtagh's cluster algorithm from Statlib

    Changes:

    the one-dimensional upper-triangular diss() array is replaced
    with the lower triangular diss[][] which includes the zero diagonal
		
    nodeNum[i] is set to zero to indicate node i has merged
		
    but the numbers returned in ia[] and ib[] are the same as before

    a bug affecting the two centroid methods has been fixed

    jmb
  */

  /*		
		c++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++c
		c                                                            c
		c  hierarchical clustering using (user-specified) criterion. c
		c                                                            c
		c  parameters:                                               c
		c                                                            c
		c  diss[][]          dissimilarities in lower half diagonal  c
		c                    storage; len = n.(n+1)/2,               c
		c  iopt              clustering criterion to be used,        c
		c  ia, ib, crit      history of agglomerations; dimensions   c
		c                    n, first n-1 locations only used,       c
		c  membr, nn, disnn  vectors of length n, used to store      c 
		c                    cluster cardinalities, current nearest  c
		c                    neighbour, and the dissimilarity assoc. c
		c                    with the latter.                        c
		c  flag              boolean indicator of agglomerable obj./ c
		c                    clusters.                               c
		c                                                            c
		c  f. murtagh, esa/eso/stecf, garching, february 1986.       c
		c                                                            c
		c------------------------------------------------------------c
  */

  /*
    c
    c  initializations
    c
  */
  try
    {
      int n = diss.length;
      int nodeNum[] = new int [n];			// the node number of this entity (negative if an interior node, zero if merged )
      int membr[] = new int [n];				// number of members in each entity
      int nn[] = new int [n];					// nearest neighbour to each entity
      double disnn[] = new double [n];		// distance to the nearest neighbour
			
      pb.SetTotalCount (n * n / 2);

      for (int i=0; i < n; ++i)
	{
	  membr [i] = 1;
	  nodeNum[i] = i+1;
	}


      int ncl = n;
      /*
	c
	c  carry out an agglomeration - first create list of nearest neighbours
	c
      */
      double dmin;
      int im = 0, jm = 0;
      int i2, j2, jj = 0;

      for (int i=1; i < n; ++i)
	{
	  dmin = Double.MAX_VALUE;
	  for (int j=0; j < i; ++j)
	    {
	      if (diss[i][j] < dmin)
		{
		  dmin = diss[i][j];
		  jm = j;
		}
	    }
	  nn [i] = jm;
	  disnn [i] = dmin;
	}
		
      do		// main clustering loop
	{

	  /*
	    c     next, determine least diss. using list of nearest neighbours
	  */
	  dmin = Double.MAX_VALUE;

	  for (int i=1; i < n; ++i)
	    {
	      if (nodeNum[i] != 0 && disnn[i] < dmin)
		{
		  dmin = disnn[i];
		  im = i;
		  jm = nn[i];
		}
	    }

	  /*
	    c
	    c  this allows an agglomeration to be carried out.
	    c
	  */
	  /*	it is always the case that im > jm, so the following

		i2 = Math.max (im, jm);
		j2 = Math.min (im, jm);

		is replaced by:
	  */
	  i2 = im;
	  j2 = jm;
	  ia [n-ncl] = nodeNum [i2] - (nodeNum[i2] > 0 ? 1 : 0);		// reduce positive node numbers back to zero origin
	  ib [n-ncl] = nodeNum [j2] - (nodeNum[j2] > 0 ? 1 : 0);
	  crit [n-ncl] = dmin;
	  --ncl;
	  nodeNum [i2] = - (n - ncl);		// assign negative interior node number

	  /*
	    c
	    c  update dissimilarities from new cluster.
	    c
	    combine elements i2 and j2 into a new one at index i2, and mark j2 inactive

	  */
	  nodeNum [j2] = 0;			// flag as merged
	  dmin = Double.MAX_VALUE;

	  double xx;
	  double dis1, dis2;
	  double newdis = 0.0;
	  int x;

	  for (int k=0; k < n; ++k)
	    {
	      if (nodeNum[k] != 0 && k != i2)
		{
		  dis1 = (i2 > k) ? diss[i2][k] : diss[k][i2];
		  dis2 = (j2 > k) ? diss[j2][k] : diss[k][j2];
		  xx = (i2 > j2) ? diss[i2][j2] : diss[j2][i2];
		  switch (iopt)
		    {


		    case SINGLE_LINKAGE:
		      newdis = Math.min (dis1, dis2);
		      break;

		    case COMPLETE_LINKAGE:
		      newdis = Math.max (dis1, dis2);
		      break;

		    case UNWEIGHTED_ARITHMETIC_AVERAGE:
		      newdis = (membr[i2] * dis1 + membr[j2] * dis2) / (membr[i2] + membr[j2]);
		      break;

		    case WEIGHTED_ARITHMETIC_AVERAGE:
		      newdis = (dis1 + dis2) / 2;
		      break;

		    case UNWEIGHTED_CENTROID:
		      int tot = membr[i2] + membr[j2];
		      newdis = (membr[i2]*dis1 + membr[j2]*dis2 - membr[i2] * membr[j2] * xx / tot) / tot;
		      break;

		    case WEIGHTED_CENTROID:
		      newdis = (dis1 + dis2) / 2 - xx / 4;
		      break;

		    case WARDS_MINIMUM_VARIANCE:
		      newdis = ((membr[i2] + membr[k]) * dis1 + (membr[j2] + membr[k]) * dis2 - membr[k] * xx) / (membr[i2] + membr[j2] + membr[k]);
		      break;
		    }

		  if (i2 > k)
		    {
		      diss[i2][k] = newdis;
		      if (newdis < dmin)		// find nearest neighbour of i2
			{
			  dmin = newdis;
			  jj = k;
			}
		    }
		  else
		    {
		      diss[k][i2] = newdis;

		      // FIX: the rest of the else clause is a fix by JMB to ensure
		      // correct nearest neighbours are found when a non-monotone
		      // clustering method (e.g. the centroid methods) are used

		      if (newdis < disnn[k])		// find nearest neighbour of i2
			{
			  disnn[k] = newdis;
			  nn[k] = i2;
			}
		    }
		}
	    }
	  membr[i2] += membr[j2];		// add in element count from j2
	  disnn[i2] = dmin;
	  nn[i2] = jj;

	  /*
	    c
	    c  update list of nns insofar as this is required.
	    c
	  */

	  for (int i=1; i < n; ++i)
	    {
	      if (nodeNum[i] != 0 && (nn[i] == i2 || nn[i] == j2))
		{
		  dmin = Double.MAX_VALUE;
		  for (int j=0; j < i; ++j)
		    {
		      if (nodeNum[j] != 0 && i != j && diss[i][j] < dmin)
			{
			  dmin = diss[i][j];
			  jj = j;
			}
		    }
		  nn[i] = jj;
		  disnn[i] = dmin;
		}
	    }

	  pb.ShowProgress (ncl);
	  /*
	    keep agglomerating until finished
	  */
	} while (ncl > 1);

      return;
    }
  catch (Exception e)
    {
      throw new Exception ("Problem doing clustering: " + e.getMessage());
    }
}


public String[][] getParameterInfo ()
{
  String[][] info = 
  {
    {"inputFormat", "string", "format of input data"},
    {"nameLocation", "string", "location of sample names"},
    {"distanceMeasure", "string", "method for calculating inter-sample distances"},
    {"clusterMethod", "string", "clustering algorithm to be used"},
    {"sampleNames", "list of words separated by spaces and/or commas", "names of the samples, if any"},
    {"sampleData", "list of numbers and words separated by spaces and/or commas", "data for each sample"},
  };
  return info;
}

public void init()
{
  // TODO: Place additional initialization code here
  String s;
		
  s = this.getParameter ("inputFormat");
  if (s != null)
    chInputFormat.select (s);

  s = this.getParameter ("nameLocation");
  if (s != null)
    chNameLocation.select (s);
		
  s = this.getParameter ("clusterMethod");
  if (s != null)
    chClusterMethod.select (s);
		
  s = this.getParameter ("distanceMeasure");
  if (s != null)
    chDistanceMeasure.select (s);
		
  s = this.getParameter ("sampleNames");
  if (s != null)
    nibSampleNames.setText (s.replace (',', '\n'));

  s = this.getParameter ("sampleData");
  if (s != null)
    nibSampleData.setText (s.replace (',', '\n'));

}

public float [][] CharMatrixFromText (ProgressBar pb, String s, int numSamples, int numChars, String names[], boolean getNames, boolean hasCharNames, boolean transpose, boolean emptyCells) throws Exception
{
  pb.SetTotalCount (numSamples * numChars);
		
  // use only floats to store characters for computing distances

  float values[][] = new float[numSamples][numChars];
  StringTokenizer tok;
  if (emptyCells)
    tok = new StringTokenizer (s, "\t \r\n", true);	// keep tabs significant
  else
    tok = new StringTokenizer (s);

  tok.nextToken ();		// skip the number of characters, which we know already
		
  if (emptyCells)
    tok.nextToken ();	// skip the end of line assumed to follow

  if (emptyCells && hasCharNames && getNames)
    tok.nextToken ();  // skip the extra TAB at the top left of the table
  try				// protect block in case of bad numbers
    {
			
      String tk;
      if (transpose)
	{
	  if (getNames)		// get sample names if necessary
	    {
	      for (int i=0; i < numSamples; ++i)
		{
		  names[i] = tok.nextToken ();
		  if (emptyCells)
		    tok.nextToken ();		// throw away delimiter
		}
	    }
	  for (int j=0; j < numChars; ++j)
	    {
	      if (hasCharNames)
		tok.nextToken (); // skip the character name
	      if (emptyCells)
		{
		  for (int i=0; i < numSamples; ++i)
		    {
		      tk = tok.nextToken ();
		      if (! Character.isSpace (tk.charAt(0)))
			{
			  values[i][j] = Float.valueOf (tk) .floatValue ();
			  tok.nextToken ();	// skip the delimiter
			}
		    }
		}
	      else
		{
		  for (int i=0; i < numSamples; ++i)
		    values[i][j] = Float.valueOf (tok.nextToken()) .floatValue ();
		}
	      pb.ShowProgress (numSamples);
	    }
	}
      else
	{
	  if (hasCharNames)
	    for (int j=0; j < numChars; ++j) // skip character names
	      tok.nextToken ();

	  for (int i=0; i < numSamples; ++i)
	    {
	      if (getNames)
		{
		  names[i] = tok.nextToken ();
		  if (emptyCells)
		    tok.nextToken ();		// discard tab
		}

	      if (emptyCells)
		{
		  for (int j=0; j < numChars; ++j)
		    {
		      tk = tok.nextToken ();
		      if (! Character.isSpace (tk.charAt(0)))
			{
			  values[i][j] = Float.valueOf (tk) .floatValue ();
			  tok.nextToken ();	// skip the delimiter
			}
		    }
		}
	      else
		{
		  for (int j=0; j < numChars; ++j)
		    values[i][j] = Float.valueOf (tok.nextToken()) .floatValue ();
		}
	      pb.ShowProgress (numChars);
	    }
	}

    }
  catch (Exception e)
    {
      throw new Exception ("Trouble getting character value from data window: " + e.getMessage () );
    }
  return values;
}


public float[][] resampleJacknife (float [][]values, int dropChar)
{
  // duplicate the array values, but delete column dropChar, where 0 <= dropChar < nc

  int ns = values.length; // number of samples
  int nc = values[0].length; // number of characters
  float [][] newValues = new float[ns][nc-1];
  
  for (int s = 0; s < ns; ++s)
  {
    System.arraycopy (values[s], 0, newValues[s], 0, dropChar);
    System.arraycopy (values[s], dropChar + 1, newValues[s], dropChar, nc - 1 - dropChar);
  }
  return newValues;
}

public float[][] resamplePermute (float [][]values, Ranecu rng, int [] permutation)
{
  // permute the samples; i.e. resample the rows
  // of the matrix without replacement
  // the permutation used is returned in 'permutation'
  // ie. newvalues[i] = values[permutation[i]]

  int ns = values.length; // number of samples
  int nc = values[0].length; // number of characters
  float [][] newValues = new float[ns][]; // this will just hold a row-wise permutation of the values array
  int [] sampleIndicies = new int[ns];
  int tmp;

  // make a permutation of the numbers 0, 1, ..., ns-1
  rng.raw (sampleIndicies, ns, ns, -1);
  for (int i=0; i < ns; ++i)
    sampleIndicies[i] += i;

  for (int i=0; i < ns; ++i)
    newValues[i] = values[i];

  for (int i=0; i < ns; ++i)
    permutation[i] = i;

  for (int i=0; i < ns - 1; ++i)
    if (sampleIndicies[i] != i)
      {
	tmp = permutation[i];
	permutation[i] = permutation[sampleIndicies[i]];
	permutation[sampleIndicies[i]] = tmp;
      }

  for (int i=0; i < ns; ++i)
    newValues[i] = values[permutation[i]];

  return newValues;
}

public float[][] resampleBootstrap (float [][]values, Ranecu rng)
{
  // resample the character values with replacement
  // this is done simultaneously across samples
  // i.e. we are just resampling the columns of the matrix
  // with replacement

  int ns = values.length; // number of samples
  int nc = values[0].length; // number of characters
  float [][] newValues = new float[ns][nc];
  int [] charIndicies = new int[nc];

  rng.raw (charIndicies, nc, nc);

  for (int i=0; i < ns; ++i)
    for (int j=0; j < nc; ++j)
      newValues[i][j] = values[i][charIndicies[j]];

  return newValues;
}

public void shellSort (int [] a)
{
  int n = a.length;
  int gap = 1;
  // find the first gap, using the sequence:
  // g(1) = 1, g(i+1) = 3 * g(i) + 1
  // stop with g(i) when g(i+2) >= n
  // as per Knuth, Searching & Sorting (p. 95)

  while (gap < n)
    gap += (gap << 2) + 1;
  
  // get g(i) given g(i+2)

  gap = (gap - 1) / 3;
  gap = (gap - 1) / 3;

  int i;

  while (gap >= 1)
    {
      for (int j = gap; j < n; ++j)
	{ 
	  int k = a[j];
	  for (i = j - gap; i >= 0; i -= gap)
	  {
	    if (k >= a[i])
	      {
		break;
	      }
	    a[i+gap] = a[i];
	  }
	  a[i+gap] = k;
	}
      gap = (gap - 1) / 3;
    }
}
	    


  
public float[][] resampleMultinomial (float [][]values, Ranecu rng)
{
  int ns = values.length; // number of samples
  int nc = values[0].length; // number of characters
  float [][] newValues = new float[ns][nc];

  for (int s=0; s < ns; ++s)
    {
      // let m be the floor of (0.5 + sum of characters) 
      // we generate m random integers between 0 and m-1 inclusive,
      // and essentially just use the inverse of the cumulative distribution
      // function of observed counts to thereby get m multinomially distributed
      // draws on nc characters

      // here we generate these s quantiles

      int sum = (int) charSums[s];
      int [] quantiles = new int[sum];
      rng.raw(quantiles, sum, sum);
      
      //  sort them

      shellSort (quantiles);

      // work through the newValues array, incrementing 
      // the appropriate character slots

      int j = 0;  // counter up to sum
      int c = 0;  // character slot
      double p = values[s][0];		// cumulative character sum

      while (j < sum)
	{
	  if (quantiles[j] <= p)
	    {
	      ++newValues[s][c];
	      ++j;
	    }
	  else
	    {
	      ++c;
	      p += values[s][c];
	    }
	}
    }      

  return newValues;
}

public int poissonCutoff (double mu, double thresh)
{
  /*
    return the value of N for which the tail
    of the poisson distribution with mean mu
    is less than thresh
  */

  int N = 0;
  double p = Math.exp (-mu);
  double q = 1;
  while (q > thresh)
    {
      q -= p;
      ++N;
      p = p * mu / N;
    }
  return N;
}

/* a straightforward method based directly on the
   density function - just add up terms in the series
   for exp(mu) until they exceed p exp(mu), where
   p is assumed uniform(0,1).

   In case of p = 1, we bail with
   the given cutoff.  The main savings is that only
   one uniform random number is used per poisson variate.

   For the test in this file, the method is as fast as Knuth's
   for mu=1, twice as fast for mu=10, and six times as fast
   for mu=100.
*/

public int randPoisson (double p, double mu, int cutoff)
{
  double x = 1.0;
  double y = 1.0;
  double cut = p * Math.exp(mu);

  int i = 0;

  if (p == 1.0)
    return cutoff;

  while (x < cut && i < cutoff)
    {
      y *= mu / (++i);
      x += y;
    }
  return i;
}

public float[][] resamplePoisson (float [][]values, Ranecu rng, int cutoff)
{
  // for each character value in each sample
  // generate a new value from a poisson distribution
  // whose mean is the old value

  int ns = values.length; // number of samples
  int nc = values[0].length; // number of characters
  float [][] newValues = new float[ns][nc];
  double [] randChars = new double [nc];

  for (int s = 0; s < ns; ++s)
    {
      // generate uniform random deviates
      rng.raw(randChars, nc);
      for (int c = 0; c < nc; ++c)
	newValues[s][c] = randPoisson (randChars[c], values[s][c], cutoff);
    }
  return newValues;
}

int[] oLeaves; // list of leaf node numbers of old tree, tree-sorted
int[] oFirst; // for each inode of old tree, the index of the 
				// first node in oLeaves that is a descendent of this inode 

int[] oLast; // for each inode of old tree, the index of the
				// last node in oLeaves that is a descendent of this inode
int[] oHashValues; // for each inode of old tree, the hash value of the subtree rooted there
int[] oSortedInode;	// the permutation giving hash values in sorted order
				// i.e. oHashValues[oSortedInode[n]] is the n-th largest hash value
int[] oLeafPos; // for the inverse permutation of the leaves

ELTree[] oInodes; // the actual array of inodes in the old tree - needed when comparing
				// partitions rooted at these nodes, because we'll need access to subtrees
double charSums[]; // array of sums of character values for each sample - used in multinomial
				// resampling, and could be used in zero-avoidance.

public void seekSupport (ELTree oTree, ELTree nTree, int [] supportCounts, int supportType)
{				// seekSupport
  // for each interior node of oTree, see whether the partition it induces among the leaf nodes
  // is also present in nTree.  If so, increment the corresponding element of supportCounts
  
  // the algorithm is simple:  hash each subtree of each tree (using the node numbers of its leaf nodes)
  // for any pair of subtrees (one in the old tree, one in the new) that have the same hash value
  // compare the partitions they induce by checking whether their left and right subtrees both
  // have the same sets of leaf nodes.

  // Currently, the hashing is reduced to mod n to allow for (fast) radix sorting.
  // This means we'd expect, on average, to compare n pairs of subtrees in detail,
  // if hashing is reasonably random.  If an average subtree has ln(n) leaf nodes,
  // then this algorithm will average n ln(n) running time

  // Caveat:  it's frequently the case that clustering generates very linear trees,
  // in which case mean subtree size is more like O(n), and so running time will
  // be more like O(n^2).  On the other hand, we do a quick check for whether two
  // subtrees have the same number of leaf nodes, and this should substantially improve
  // things even in this case - the more uniform the subtree length distribution, the
  // lower the chance of two subtrees having equal lengths.

  // If we replaced the radix sort by quicksort, leaving hash values as 32 bit integers,
  // that would force sorting time up to O(n ln n) with
  // a worse constant because of relatively expensive Java method calls,
  // but presumably radically reduce the number of equal hash values, and hence (slow) subtree comparisons


  int numLeaves = oTree.numLeafNodes;

  // in other words, the node numbers of leaf nodes descended from interior node i
  // are stored in oLeaves[oFirst[i]..oLast[i]]
  // we've made these static so they are only computed the first time this function is called

  // now the same things for the new tree

  int[] nLeaves = null;
  int[] nFirst = null;
  int[] nLast = null;
  int[] nHashValues = null;
  int[] nSortedInode = null;
  ELTree[] nInodes = null;

  // finally, we want a fast way to tell whether a given leaf node is a descendent of a given
  // inode in the new tree
  // One way to do this is to invert the permutation given by nLeaves; this goes here:

  int[] nLeafPos = null; 

  // therefore, to check whether node n is a descendent of inode i in the new tree,
  // we simply check whether nFirst[i] <= nLeafPos[n] <= nLast[i]

  // generate partition boundaries etc. for the new tree

  nTree.sortByLeaves ();
  nLeaves = new int [numLeaves];
  nTree.getLeafNodes (nLeaves, 0);

  int numInodes = numLeaves - 1;

  nFirst = new int [numInodes];
  nLast = new int [numInodes];
  nTree.getTreeExtents (nFirst, nLast, 0);
  
  nHashValues = new int [numInodes];
  nTree.hashByLeaves (nHashValues, false, supportType == SUPP_BI);

  nSortedInode = new int [numInodes];
  radixSortHashValues (nHashValues, nSortedInode);

  nInodes = new ELTree [numInodes];
  nTree.getInodes (nInodes);


  // for fast lookup, invert the permutation given by nLeaves

  nLeafPos = new int [numLeaves];
  for (int i=0; i < numLeaves; ++i)
    nLeafPos[nLeaves[i]] = i;

  // for each inode in the original tree, see if any inode
  // in the new tree has the same hash value, and if so,
  // compare the partitions induced by those inodes
  // (NB:  using the SortedInode permutation makes this a comparison of two sorted lists)

  int ni = 0;
  int oi = 0;
  int ohv = oHashValues[oSortedInode[oi]];
  int nhv = nHashValues[nSortedInode[ni]];

  while (oi < numInodes && ni < numInodes)
    {
      if (ohv < nhv)
	{
	  ++ oi;
	  if (oi == numInodes)
	    break;
	  ohv = oHashValues[oSortedInode [oi]];
	}
      else if (ohv > nhv)
	{
	  ++ ni;
	  if (ni == numInodes)
	    break;
	  nhv = nHashValues[nSortedInode [ni]];
	}
      else
	{
	  // we've found an inode in the new tree
	  // that has the same hash value as an inode in the
	  // old tree.  We will loop over all inodes in
	  // the new tree with this same hash value, and compare
	  // their partitions to that of the inode in the old tree
	  // quitting if we find one that's the same

	  int tmpni = ni;	// we will look at all new inodes with the same hash value

	same_hash_value_loop:
	  do
	    {

	      // compare the partitions induced by inode oin in the old tree
	      // and the partition induced by inode nin in the new tree
	      ELTree oin = oInodes [oSortedInode [oi]];
	      ELTree nin = nInodes [nSortedInode [tmpni]];

	      if (supportType == SUPP_BI)
		{
		  // compare the sets of leaf nodes for the old and new inodes

		  if (haveSameLeaves (oin, nin, oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      ++ supportCounts [oSortedInode[oi]];
		      break same_hash_value_loop;
		    }
		} 
	      else
		{ 
		  // supportType == SUPP_TRI - treat the inode as partitioning its descendents

		  // because the trees are both in leaf-sorted order,
		  // the partitions induced by oin and nin are equal if and only if:
		  // - their left subtrees have the same leaf nodes
		  //     and
		  // - their right subtrees have the same leaf nodes
	  
	      // in these rooted trees, inodes always have two children
	      
	      // grab the inode numbers for the four subtrees

		  if (   haveSameLeaves (oin.T[0], nin.T[0], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
		      && haveSameLeaves (oin.T[1], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      // we've found an inode in the new tree that partitions the leaves
		      // the same way as does inode oi in the old tree
		      ++ supportCounts [oSortedInode[oi]];

		      // we're ready to search for the next old inode partition, so break out of
		      // this comparison loop

		      break same_hash_value_loop;
		    }
		}
	      
	      ++ tmpni;

	    }  // same_hash_value_loop
	  // keep looping as long as there remain inodes in the new tree with the same hash value

	  while (tmpni < numInodes && nHashValues[nSortedInode[tmpni]] == ohv);

	  // a partition can only appear once in a tree, so bump up the old tree inode index
	  ++ oi;
	  if (oi == numInodes)
	    break;

	  ohv = oHashValues[oSortedInode [oi]];
	}
    }  
}

public void seekSupportUnrooted (ELTree oTree, ELTree nTree, int [] supportCounts, int supportType)
{				// seekSupportUnrooted

  // Two versions, depending on support type:
  // if supportType == SUPP_BI:
  //   for each non-trivial edge of oTree, see whether the partition its removal induces among the leaf nodes
  //   is also present in nTree.  "present" means there is an edge in nTree that induces the same partition
  //   If so, increment the corresponding element of supportCounts
  //   (an edge is non-trivial if it joins two interior nodes; there are n-3 non-trivial edges in an unrooted
  //   tree on n leaf nodes when each inode has degree 3.
  // if supportType == SUPP_TRI:
  //   for each inode of oTree, see whether the partition its removal induces among the leaf nodes
  //   is also present in nTree.  "present" means there is an inode in nTree that induces the same
  //   partition.  If so, increment the corresponding element of supportCounts (the element corresponding
  //   to the inode that is joined to its parent by that edge in the rooted representation of the tree)
  
  int numLeaves = oTree.numLeafNodes;

  // in other words, the node numbers of leaf nodes descended from interior node i
  // are stored in oLeaves[oFirst[i]..oLast[i]]
  // we've made these static so they are only computed the first time this function is called

  // now the same things for the new tree

  int[] nLeaves = null;
  int[] nFirst = null;
  int[] nLast = null;
  int[] nHashValues = null;
  int[] nSortedInode = null;
  ELTree[] nInodes = null;

  // finally, we want a fast way to tell whether a given leaf node is a descendent of a given
  // inode in the new tree
  // One way to do this is to invert the permutation given by nLeaves; this goes here:

  int[] nLeafPos = null; 

  // therefore, to check whether node n is a descendent of inode i in the new tree,
  // we simply check whether nFirst[i] <= nLeafPos[n] <= nLast[i]

  // generate partition boundaries etc. for the new tree

  nTree.sortByLeaves ();
  nLeaves = new int [numLeaves];
  nTree.getLeafNodes (nLeaves, 0);

  int numInodes = numLeaves - 2;
  nFirst = new int [numInodes];
  nLast = new int [numInodes];
  nTree.getTreeExtents (nFirst, nLast, 0);
  
  nHashValues = new int [numInodes];
  nTree.hashByLeaves (nHashValues, true, supportType == SUPP_BI);

  nSortedInode = new int [numInodes];
  radixSortHashValues (nHashValues, nSortedInode);

  nInodes = new ELTree [numInodes];
  nTree.getInodes (nInodes);

  // for fast lookup, invert the permutation given by nLeaves

  nLeafPos = new int [numLeaves];
  for (int i=0; i < numLeaves; ++i)
    nLeafPos[nLeaves[i]] = i;

  // for each inode in the original tree, see if any inode
  // in the new tree has the same hash value, and if so,
  // compare the appropriate sets of leaf nodes for equality
  // (NB:  using the SortedInode permutation makes this a comparison of two sorted lists)

  int ni = 0;
  int oi = 0;
  int ohv = oHashValues[oSortedInode[oi]];
  int nhv = nHashValues[nSortedInode[ni]];

  while (oi < numInodes && ni < numInodes)
    {
      if (ohv < nhv)
	{
	  ++ oi;
	  if (oi == numInodes)
	    break;
	  ohv = oHashValues[oSortedInode [oi]];
	}
      else if (ohv > nhv)
	{
	  ++ ni;
	  if (ni == numInodes)
	    break;
	  nhv = nHashValues[nSortedInode [ni]];
	}
      else
	{
	  // we've found an inode in the new tree
	  // that has the same hash value as an inode in the
	  // old tree.  We will loop over all inodes in
	  // the new tree with this same hash value, and compare
	  // their partitions to that of the inode in the old tree
	  // quitting if we find one that's the same

	  int tmpni = ni;	// we will look at all new inodes with the same hash value

	same_hash_value_loop:
	  do
	    {

	      // compare the partitions induced by inode oin in the old tree
	      // and the partition induced by inode nin in the new tree
	      ELTree oin = oInodes [oSortedInode [oi]];
	      ELTree nin = nInodes [nSortedInode [tmpni]];

	      if (supportType == SUPP_BI)
		{
		  // compare the sets of leaf nodes for the old and new inodes

		  // if the two inodes have either the same or complementary sets of descendents,
		  // then the partitions induced by deleting the edge from the inodes to their
		  // respective parents are the same

		  if (  haveSameLeaves (oin, nin, oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			|| ( (oin.numLeafNodes + nin.numLeafNodes == numLeaves) 
			   && haveDistinctLeaves (oin, nin, oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			     ) )
		    {
		      ++ supportCounts [oSortedInode[oi]];
		      break same_hash_value_loop;
		    }
		} 
	      else 
		{ 
		  /*

		   supportType == SUPP_TRI - treat the inode as partitioning leaves into three groups,
		   namely: leaves of left subtree, leaves of right subtree, all other leaves in tree
		   we have four sets A, B, C, and D, all subsets of S
		   and must decide whether the partitions
		   {A, B, S-(A+B)} and {C, D, S-(C+D)} are equal
		   these are the 6 possible ways:
		    1) A=C, B=D
		    2) A=C, D=(S-(A+B)) i.e. D&B=0 and |A|+|B|+|D| = |S| (we know D&C=0)
		    3) A=D, B=C
		    4) A=D, C=(S-(A+B)) i.e. C&B=0 and |A|+|B|+|C| = |S|
		    5) B=C, A=(S-(C+D)) i.e. A&D=0 and |A|+|B|+|D| = |S|
		    6) B=D, A=(S-(C+D)) i.e. A&C=0 and |A|+|B|+|C| = |S|
		   but notice that 3 is impossible because A < B and C < D (i.e. their smallest leaf nodes
		   are related this way)

		   We restructure this as

		   guess:
		   {
		     if A=C
		       {
                         if !(B=D || D=(S-(A+B)))
			   break guess;
                       }
		     else if A=D
                       {
                         if !(C=(S-(A+B)))
			   break guess;
                       }
                     else if B=C
		      {
		         if !(A=(S-(C+D)))
                           break guess;
                      }
                     else if B=D
                      {
                         if !(A=(S-(C+D)))
                           break guess;
                      }
                     else
                       break guess;
                     
		       ... partitions are equivalent ...
                   }

		  */

		guess: {
		  
		  if ( haveSameLeaves (oin.T[0], nin.T[0], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      if (!(haveSameLeaves (oin.T[1], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			    || ( (oin.T[0].numLeafNodes + oin.T[1].numLeafNodes + nin.T[1].numLeafNodes == numLeaves)
				 && haveDistinctLeaves (oin.T[1], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
				 )))
			break guess;
		    }
		  else if ( haveSameLeaves (oin.T[0], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      if (!( (oin.T[0].numLeafNodes + oin.T[1].numLeafNodes + nin.T[0].numLeafNodes == numLeaves)
			     && haveDistinctLeaves (oin.T[1], nin.T[0], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			     ))
			break guess;
		    }
		  else if (haveSameLeaves (oin.T[1], nin.T[0], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      if (!( (oin.T[0].numLeafNodes + oin.T[1].numLeafNodes + nin.T[1].numLeafNodes == numLeaves)
			     && haveDistinctLeaves (oin.T[0], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			     ))
			break guess;
		    }
		  else if (haveSameLeaves (oin.T[1], nin.T[1], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos))
		    {
		      if (!( (oin.T[0].numLeafNodes + oin.T[1].numLeafNodes + nin.T[0].numLeafNodes == numLeaves)
			     && haveDistinctLeaves (oin.T[0], nin.T[0], oLeaves, oFirst, oLast, nFirst, nLast, oLeafPos, nLeafPos)
			     ))
			break guess;
		    }
		  else
		    {
		      break guess;
		    }
		  
		  // the two partitions are the same
		  
		  ++ supportCounts [oSortedInode[oi]];

		  // we're ready to search for the next old inode partition, so break out of
		  // this comparison loop

		  break same_hash_value_loop;
		};
	      };
	      ++ tmpni;
	      
	    }  // same_hash_value_loop
	  // keep looping as long as there remain inodes in the new tree with the same hash value
	  while (tmpni < numInodes && nHashValues[nSortedInode[tmpni]] == ohv);
	  
	  // a partition can only appear once in a tree, so bump up the old tree inode index
	  ++ oi;
	  if (oi == numInodes)
	    break;
	  
	  ohv = oHashValues[oSortedInode [oi]];
	}
    }  
} // seekSupportUnrooted


boolean haveSameLeaves (ELTree t1, ELTree t2, int [] leaves1, int [] firsts1, int [] lasts1, int [] firsts2, int [] lasts2, int [] leafPos1, int [] leafPos2)
{
  // determine whether the trees t1 and t2 have the same leaves, using auxilliary information for speed
  // the trick is that we use the inverse permutation of the leaf positions in the second set,
  // which means we don't actually use leaves2[].  i.e. to check whether a leaf x is in leaves2[first2..last2],
  // we just check whether first2 <= leafPos2[x] <= last2

  int n1 = (-t1.nodeNum)-1;
  int n2 = (-t2.nodeNum)-1;

  int first1, last1, first2, last2;

  // deal with the possibility that one or both trees is simply a leaf node

  if (n1 >= 0)
    {
      first1 = firsts1 [n1];
      last1 = lasts1[n1];
    }
  else
    {
      first1 = last1 = leafPos1[-(n1+1)];
    }

  if (n2 >= 0)
    {
      first2 = firsts2 [n2];
      last2 = lasts2[n2];
    }
  else
    {
      first2 = last2 = leafPos2[-(n2+1)];
    }

  if (last1 - first1 != last2 - first2)
    return false;

  for (int i = first1; i <= last1; ++i)
    {
      int pos = leafPos2[leaves1[i]];
      if (pos < first2 || pos > last2)
	// if the position of this leaf is outside the range for tree 2, then it's not
	// one of tree 2's leaves
	return false;
    }
  // all leaves in tree 1 are in tree 2, and the number of leaves is equal, 
  // so the trees have exactly the same leaves

  return true;
}

boolean haveDistinctLeaves (ELTree t1, ELTree t2, int [] leaves1, int [] firsts1, int [] lasts1, int [] firsts2, int [] lasts2, int [] leafPos1, int [] leafPos2)
{
  // determine whether the trees t1 and t2 have distinct sets of leaves,
  //  using auxilliary information for speed
  // the trick is that we use the inverse permutation of the leaf positions in the second set,
  // which means we don't actually use leaves2[].  i.e. to check whether a leaf x is in leaves2[first2..last2],
  // we just check whether first2 <= leafPos2[x] <= last2

  // if checkCompSizes is true, we also check whether the two sets partition the entire set of nodes
  // by seeing whether their sizes add up to the total

  int n1 = (-t1.nodeNum)-1;
  int n2 = (-t2.nodeNum)-1;

  int first1, last1, first2, last2;

  // deal with the possibility that one or both trees is simply a leaf node

  if (n1 >= 0)
    {
      first1 = firsts1 [n1];
      last1 = lasts1[n1];
    }
  else
    {
      first1 = last1 = leafPos1[-(n1+1)];
    }

  if (n2 >= 0)
    {
      first2 = firsts2 [n2];
      last2 = lasts2[n2];
    }
  else
    {
      first2 = last2 = leafPos2[-(n2+1)];
    }

  for (int i = first1; i <= last1; ++i)
    {
      int pos = leafPos2[leaves1[i]];
      if (pos >= first2 && pos <= last2)
	// if the position of this leaf is in the range for tree 2, then it's
	// one of tree 2's leaves, so the two sets can't be a partition of all leaves
	return false;
    }
  // no leaves in tree 1 are in tree 2
  // so the trees have distinct sets of leaves

  return true;
}

void radixSortHashValues (int [] value, int [] index)
{
  // reduce hash values to the range [0..n-1] and
  // then perform Seward's radix sort, essentially as described
  // in Knuth, Searching and Sorting, Algorithm D, p. 78
  // this gives the inode numbers in increasing order of (reduced) hash value

  int n = value.length;
  int [] count = new int[n];

  // reduce hash values to the range [0..n-1]
  // dropping bit 0  because it's always 1
  // this also guarantees a non-negative value
  // so that the modulus is non-negative

  for (int i=0; i < n; ++i)
    value[i] = (value[i] >>> 1) % n;

  // calculate frequencies of hash values

  for (int i=0; i < n; ++i)
    ++ count [value [i]];

  // calculate the number of hash values less than
  // or equal to each (i.e. just make the array
  // cumulative)

  for (int i=1; i < n; ++i)
    count[i] += count[i-1];

  // put the node numbers in order of increasing
  // hash value (there's no reason to care about stability of the sort, in this case)

  for (int i=0; i < n; ++i)
    index[--count[value[i]]] = i;
  
} 
}
