package treelib;
import java.util.*;
import mylib.*;

class TreeStringTokenizer extends StringTokenizer
{
  // a special stringtokenizer that ignores whitespace but
  // returns other delimiters as tokens

  String delims;

  public TreeStringTokenizer (String s, String delims)
  {
    super (s, delims + " \n\r\t", true);
    this.delims = delims;
  }

  public String nextToken ()
  {
    String retVal;
    do
      {
	retVal = super.nextToken ();
      } while (Character.isSpace (retVal.charAt (0)));
    return retVal;
  }
}

/*
 *
 * ELTree	- edge-labelled binary tree class
 *
 *	definition:
 *					
 *			T := node
 *				| [T1, edgeLabel1, T2, edgeLabel2, ...]
 *
 *			where T1, T2, ... are ELTrees,
 *			and node, edgeLabel1, and edgeLabel2, ... are arbitrary java Objects
 *
 * so an ELTree is either a node
 * or a pair of trees connected to an interior node with a label
 * associated with each of the two connecting edges
 * 
*/


public class ELTree		
{
  public static final String SPACES = "                                                                                     ";
  public static final String DASHES = "-------------------------------------------------------------------------------------";

  static char openSubtree  = '(';	// delimiters for parsing a string into a tree
  static char closeSubtree = ')';
  static char delimLabel   = ':';
  static char delimSubtree = ',';

  static String EOL = "\n";			// delimiters for outputting the tree as a string
  static String delim = "\t";

  static String inodeSuffix = "'";	// string to concatenate with inode number to indicate an interior node

  public int numChildren;		// number of child nodes
  public int numNodes;			// total nodes in tree rooted here
  public int numLeafNodes;		// total leaf nodes in tree rooted here
  public int numOpenLeaves;		// total leaf nodes in open part of subtree
  public int tmp;					// temporary storage, e.g. for sorting
  public int nodeNum;				// number of this node
  public double depth;			// depth of subtree: maximum distance of any descendent from the root
  public boolean open;			// true if subtrees should be accessed

  public Object node;			// label for this node

  public ELTree T[];				// the child trees of this node
	
  public Object L[];				// labels for the edges to the child trees
	
  public ELTree ()
  {
    this (2);
  }
	
  public ELTree (int childCapacity)
  {
    // construct an unfilled  node
    this.numChildren = 0;
    node = null;
    numNodes = 1;

    // if this node can have no children, it is a leaf node

    numLeafNodes = (childCapacity == 0) ? 1 : 0;

    numOpenLeaves = 0;
    T = new ELTree[childCapacity];
    L = new Object[childCapacity];
    open = true;
  }

  public ELTree (Object node)
  {
    // construct a leaf node

    numChildren = 0;
    this.node = node;
    T = null;
    L = null;
    numNodes = numLeafNodes = numOpenLeaves = 1;
    open = true;
    depth = 0.0;
  }

  static public ELTree fromTokenizer (StringTokenizer stok) throws Exception
  {
    // construct a tree of nodes from a stream of tokens


    String tok = stok.nextToken ();
    if (tok.charAt (0) != openSubtree)			// return a new leaf node
      {
	return new ELTree (tok);
      }
		
    ELTree me = new ELTree ();

    do
      {
	ELTree T =  ELTree.fromTokenizer (stok);	// parse the a subtree
		
	tok = stok.nextToken ();			// get its label, if any
	String L = "";
	if (tok.charAt (0) == delimLabel)
	  {
	    L = stok.nextToken ();
	    tok = stok.nextToken ();
	  }
	me.addChild (T, L);
      }
    while (tok.charAt (0) == delimSubtree);			// continue until no more subtrees

    if (tok.charAt (0) != closeSubtree)		// finished subtree list, should find a closing delimiter
      throw new Exception ("Bad tree syntax at " + tok + stok.nextToken () );

    return me;
  }

  public static ELTree fromString (String s)		// build a tree by parsing a string description
  {
    return fromString (s, "():,");
  }

public static ELTree fromString (String s, String delims)		// build a tree by parsing a string description
{
  // delims is a string with 4 characters:
  // 1: open sub tree 
  // 2: close sub tree
  // 3: separate sub tree from edge label
  // 4: separate sub tree 1 from sub tree 2
		
  // create a tokenizer that returns tokens, nodelabels, and edgelabels

  ELTree.setDelims (delims);

  TreeStringTokenizer stok = new TreeStringTokenizer (s, delims);

  ELTree retVal;
  try
    {
      retVal = fromTokenizer (stok);
    }
  catch (Exception e)
    {
      return null;
    }

  retVal.assignInteriorNames (0);
  return retVal;

}

public static ELTree fromClusterOutput (int node1[], int node2[], double clusterDistance[], String nodeName[], int samplePermutation[])		
		
		// Build a tree from clustering algorithm output
		// The arrays node1 and node2 are indexed by
		// edge number, and their values are node numbers.
		// Negative node numbers represent interior nodes.
		// Non-negative node numbers represent leaf nodes, whose names
		// are in nodeNames[] 

// clusterDistance is an array giving the distance at which cluster
// combining occurred for each step in the clustering
// We convert these to distances
// between clustered nodes and the newly created Inode by
// remembering the "depth" of each node as we build the tree

// Note:  interior node numbers must occur in node1[] and node2[]
// in increasing order of magnitude for edge labels to be correct.

// We construct an ELTree for each interior and leaf,
// then assign children to parents.
{
  int numInodes = node1.length;
  int numLeaves = nodeName.length;
  int numNodes = numInodes + numLeaves;
  double depth[] = new double [numNodes];		// elements are initialized to zero
  ELTree nodes[] = new ELTree [numNodes];

  int i;

  // allocate and name interior nodes

  for (i=0; i < numInodes; ++i)
    {
      nodes[i] = new ELTree (2);
      nodes[i].node = String.valueOf (numInodes - i) + inodeSuffix;
      nodes[i].nodeNum = -(numInodes - i);
    }


  // allocate leaf nodes
  // if the samples have been permuted (so that the node currently numbered i
  // was originally numbered samplePermutation[i],
  // then assign the original node numbers

  for (; i < numNodes; ++i)
    {
      nodes[i] = new ELTree (0);
      int nodeNum = i - numInodes;
      if (samplePermutation != null)
	nodeNum = samplePermutation[nodeNum];

      nodes[i].node = nodeName[nodeNum];
      nodes[i].nodeNum = nodeNum;
    }

  // for each interior node, connect its children,
  // labelling the edges with the appropriate distances

  for (i=0; i < numInodes; ++i)
    {
      int j = numInodes - i - 1;
      depth[j] = clusterDistance[i] / 2.0;
      nodes[j].addChild (nodes [numInodes + node1[i]], String.valueOf (depth[j] - depth[numInodes + node1[i]]));
      nodes[j].addChild (nodes [numInodes + node2[i]], String.valueOf (depth[j] - depth[numInodes + node2[i]]));
    }

  // the complete ELTree is now rooted at nodes[numInodes-1]

  return nodes[0];
}

  public static ELTree fromNeighbourOutput (int node1[], int node2[], double mergeDist1[], double mergeDist2[], String nodeName[])		
		
		// Build a tree from neighbour joining algorithm output
		// The arrays node1 and node2 are indexed by
		// edge number, and their values are node numbers.
		// Negative node numbers represent interior nodes.
		// Non-negative node numbers represent leaf nodes, whose names
		// are in nodeNames[] 

// mergeDist1[] and mergeDist2[] are the computed branch
// lengths between nodes merged at each step, and the new
// node representing their merger

// We construct an ELTree for each interior and leaf,
// then assign children to parents.
{
  int numInodes = node1.length - 1;
  int numLeaves = nodeName.length;
  int numNodes = numInodes + numLeaves;
  ELTree nodes[] = new ELTree [numNodes];

  int i;

  // allocate and name interior nodes

  for (i=0; i < numInodes; ++i)
    {
      nodes[i] = new ELTree (2);
      nodes[i].node = String.valueOf (numInodes - i) + inodeSuffix;
      nodes[i].nodeNum = - (numInodes - i);
    }

  // allocate leaf nodes

  for (; i < numNodes; ++i)
    {
      nodes[i] = new ELTree (0);
      nodes[i].node = nodeName[i - numInodes];
      nodes[i].nodeNum = i - numInodes;
    }

  // for each interior node but the last, connect its children,
  // labelling the edges with the appropriate distances

  int j;
  for (i=0; i < numInodes; ++i)
    {
      j = numInodes - i - 1;
      nodes[j].addChild (nodes [numInodes + node1[i]], String.valueOf (mergeDist1[i]));
      nodes[j].addChild (nodes [numInodes + node2[i]], String.valueOf (mergeDist2[i]));
    }

  // the last merge involves adding a third child to the root

  if (node1[i] == -numInodes)
    j = node2[i];
  else
    j = node1[i];

  nodes[0].addChild (nodes [j + numInodes], String.valueOf (mergeDist1[i]));

  // the complete ELTree is now rooted at nodes[0]
		
  return nodes[0];
}

public void addChild (ELTree cT, Object cL)
{
  // add a child node
  if (numChildren == T.length)		// expand child array if necessary
    {
      ELTree tmpT[] = T;
      Object tmpL[] = L;
      T = new ELTree [2 * numChildren];
      L = new Object [2 * numChildren];
      System.arraycopy (tmpT, 0, T, 0, numChildren);
      System.arraycopy (tmpL, 0, L, 0, numChildren);
      tmpT = null;
      tmpL = null;
    }
  T [numChildren] = cT;
  L [numChildren] = cL;
  ++ numChildren;
  numNodes += cT.numNodes;
  numLeafNodes += cT.numLeafNodes;
  numOpenLeaves += cT.numOpenLeaves;
}

public void swapChildren (int c1, int c2)
{
  ELTree tmpE = T[c1];
  T[c1] = T[c2];
  T[c2] = tmpE;

  Object tmpO = L[c1];
  L[c1] = L[c2];
  L[c2] = tmpO;
}

public static void setDelims (String delims)
{
  if (delims.length () != 4)
    return;

  openSubtree = delims.charAt (0);
  closeSubtree = delims.charAt (1);
  delimLabel = delims.charAt (2);
  delimSubtree = delims.charAt (3);
}

public String toString ()
{
  return toString (true);
}

public String toString (boolean dontHideSubtrees)
{
  // dump out the tree in infix order,
  // considering all closed nodes to be leaf nodes
  // this is actually depth-first order, because
  // we don't output labels for interior nodes

  // if dontHideSubtrees is true, the entire tree is
  // output, regardless of the open/closed state of nodes
  // otherwise, the subtrees of any closed node are not examined

  if (numChildren == 0 || ! (open || dontHideSubtrees))
    {
      return node.toString ();
    }
  else
    {
      StringBuffer buf = new StringBuffer (numNodes * 10);
      buf.append (openSubtree);
      for (int i=0; i < numChildren; ++i)
	{
	  if (i > 0)
	    buf.append (delimSubtree);

	  buf.append (T[i].toString (dontHideSubtrees));
	  if (L[i] != null)
	    buf.append (delimLabel).append (L[i].toString ());
	}
      buf.append (closeSubtree);
      return buf.toString ();
    }
}

/*

  OLD version - shows leaf nodes, makes huge wide lists 
  public String toTreeList ()
  {
  // dump out the tree in infix order,
  // considering all closed nodes to be leaf nodes
  // Every node is output on one line, and
  // a node is preceded by a number of spaces equal
  // to its depth in the tree
  // interior nodes are preceded by a + or - to indicate
  // whether they are open or closed

  return toTreeList (0);
  }

  private String getFullLabel ()
  {
  String s;
  if (numChildren == 0)
  {
  s = " Leaf Node: " + node.toString () + EOL;
  }
  else
  {
  s = (open ? "-" : "+") + node.toString ();
  s += "  has " + numChildren + " children and " + numLeafNodes + " descendents." + EOL;
  }
  return s;
  }

  private String toTreeList (int level)
  {
  int numsp = 2 * level + 2 - ((level >= 9) ? 1 : 0) - ((level >= 99) ? 1 : 0);
  if (numChildren == 0 || !open)
  {
  return (level+1) + ":" + SPACES.substring (0, numsp) + getFullLabel ();
  }
  else
  {
  StringBuffer buf = new StringBuffer (numNodes * 10);
  buf .append (level+1) .append (':') .append (SPACES.substring (0, numsp));
  buf.append ( getFullLabel ());
  ++ level;
  for (int i=0; i < numChildren; ++i)
  {
  buf.append (T[i].toTreeList (level));
  }
  return buf.toString ();
  }
  }

*/
public String toTreeList ()
{
  // dump out the tree in infix order,
  // considering all closed nodes to be leaf nodes
  // Every node is output on one line, and
  // a node is preceded by a number of spaces equal
  // to its depth in the tree
  // interior nodes are preceded by a + or - to indicate
  // whether they are open or closed

  return toTreeList (0);
}

private String getFullLabel ()
{
  String s;
  if (numChildren == 0)
    {
      s = "";
    }
  else
    {
      s = (open ? "- " : "+ ") + node.toString ();
      s += "  C=" + numChildren + " L=" + numLeafNodes  + EOL;
    }
  return s;
}

private String toTreeList (int level)
{
  if (numChildren == 0)
    return "";
  else if (!open)
    {
      return (level+1) + ": " + getFullLabel ();
    }
  else
    {
      StringBuffer buf = new StringBuffer (numNodes * 10);
      buf .append (level+1) .append (": " );
      buf.append ( getFullLabel ());
      ++ level;
      for (int i=0; i < numChildren; ++i)
	{
	  buf.append (T[i].toTreeList (level));
	}
      return buf.toString ();
    }
}


public void open ()
{
  open (true);
}

public boolean isOpen ()
{
  return open;
}

public void open (boolean o)
{
  open = o;
}

public void close ()
{
  open (false);
}

public void openAll ()
{
  openAll (true);
}

public void closeAll ()
{
  openAll (false);
}

public void openAll (boolean o)	
{
  // recursively open/close a subtree

  open = o;
  for (int i=0; i < numChildren; ++i)
    T[i].openAll (o);
}

public ELTree [] getNodeList ()
{
  // return the prefix order list of nodes in this tree
  ELTree nlist[] = new ELTree [numNodes];
  fillNodeList (this, nlist, 0);
  return nlist;
}

/* OLD VERSION - includes leaf nodes 
   private static int fillNodeList (ELTree tree, ELTree nlist[], int index)
   {
   nlist[index++] = tree;
   for (int i=0; i < tree.numChildren; ++i)
   index = fillNodeList (tree.T[i], nlist, index);
   return index;
   }
*/

  private static int fillNodeList (ELTree tree, ELTree nlist[], int index)
    {
      if (tree.numChildren > 0)
	{
	  nlist[index++] = tree;
	  for (int i=0; i < tree.numChildren; ++i)
	    index = fillNodeList (tree.T[i], nlist, index);
	}
      return index;
    }

  public ELTree getDescendent (int n)
    {
      // return the nth descendent of this node in prefix traversal order:

      //		- a node is visited before its children
      //		- child i and its subtree are visited before child j and its subtree
      //		  for all i < j

      //		only visit subtrees if this node is open
      
      //		n should be >= 0

      if (n==0)
	return this;

      if (n<0)					// the index is too small
	return null;

      if (numChildren == 0 || !open)
	return null;

      int i;
      -- n;
      ELTree retVal = null;
      for (i=0; i < numChildren; ++i)
	{
	  retVal = T[i].getDescendent (n);
	  if (retVal != null)
	    break;
	}
      return retVal;
    }

  public ELTree getVisibleDescendent (ELTree nlist[], int n)
    {
      // return the nth visible descendent in a tree

      int j = -1;
      int i = 0;
      while (i < nlist.length)
	{
	  ++j;
	  if (j == n)
	    return nlist[i];
	  if (nlist[i].open)		// if this node is open, look at its children
	    ++i;
	  else					// otherwise, skip all its descendents
	    i += nlist[i].numNodes - nlist[i].numLeafNodes;
	}
      return null;
    }

  public static double calcDepth (ELTree tree)
    {
      // calculate depths, assuming the tree has edges that can
      // be casted to double values
      // absolute values of distances are used

      return calcDepth (tree, 0.0);
    }

  static double calcDepth (ELTree tree, double d)
    {
      double maxD = d;
      if (tree.open)
	{
	  for (int i=0; i < tree.numChildren; ++i)
	    {
	      double childDepth = calcDepth (tree.T[i], d  + Math.abs (Double.valueOf( (String) tree.L[i]). doubleValue ()));
	      if (childDepth > maxD)
		maxD = childDepth;
	    }
	}
      return tree.depth = maxD;
    }

  public static int calcOpenLeaves (ELTree tree)
    {
      int nOpen = 1;
      if (tree.open && tree.numChildren > 0)
	{
	  nOpen = 0;
	  for (int i=0; i < tree.numChildren; ++i)
	    {
	      nOpen +=  calcOpenLeaves (tree.T[i]);
	    }
	}
      return tree.numOpenLeaves = nOpen;
    }

  int assignInteriorNames (int n)
    {
      for (int i=0; i < numChildren; ++i)
	{
	  n =  T[i].assignInteriorNames (n);
	}
      if (numChildren > 0)
	{
	  node = (n+1) + inodeSuffix;
	  ++n;
	}
      return n;
    }

  public String toTextPicture ()
    {
      StringBuffer sb = new StringBuffer (numNodes * 20);
      boolean stems[] = new boolean [numNodes+1];
      toTextPicture (sb, stems, 0, true);
      return sb.toString ();
    }

  private void drawStems (StringBuffer sb, boolean stems[], int depth, boolean connect)
    {
      for (int i=0; i < depth; ++i)
	{
	  if (stems[i])
	    sb.append ("|    ");
	  else
	    sb.append ("     ");
	}
      if (connect && depth >= 0)
	sb.append ("+----");
    }

  private void toTextPicture (StringBuffer sb, boolean stems[], int drawDepth, boolean lastChild)
    {
      if (numChildren == 0)
	{
	  drawStems (sb, stems, drawDepth - 1, true);
	  sb.append (node) .append (EOL);
	  stems[drawDepth-1] = ! lastChild;
	}
      else
	{
	  for (int i=0; i < numChildren; ++i)
	    {
	      T[i].toTextPicture (sb, stems, drawDepth + 1, i == numChildren - 1);

	      if (i == 0)
		if (drawDepth > 0)
		  stems[drawDepth-1] = ! lastChild;

	      if (i == (numChildren - 2) / 2)
		{
		  drawStems (sb, stems, drawDepth - 1, numChildren ==2);
		  sb.append (node) .append (EOL);
		}
	      else if (i != numChildren - 1)
		{
		  drawStems (sb, stems, drawDepth + 1, false);
		  sb.append (EOL);
		}
	    }
	  stems[drawDepth] = false;
	}
    }

  public String lengthTable ()
    {
      StringBuffer sb = new StringBuffer (numNodes * 20);
      lengthTable (sb);
      return sb.toString ();
    }

  public static void setSeparatorChars (String delim, String EOL)
    {
      ELTree.delim = delim;
      ELTree.EOL = EOL;
    }

  private void lengthTable (StringBuffer sb)
    {
      for (int i=0; i < numChildren; ++i)
	{
	  sb.append (node) .append (delim) .append ((T[i].node)) .append (delim) .append (L[i]). append (EOL);

	  if (T[i].numChildren != 0)
	    T[i].lengthTable (sb);
	}
    }

  public static String getInodeSuffix ()
    {
      return inodeSuffix;
    }

  public static void setInodeSuffix (String ip)
    {
      inodeSuffix = ip;
    }


  public void sortByLeaves ()
    {
      /* sort children by increasing order of their smallest leaf node */

      int i, j;

      if (numChildren == 0)
	{
	  tmp = nodeNum;
	}
      else
	{
	  for (i=0; i < numChildren; ++i)
	    T[i].sortByLeaves ();	

	  /* 
	     each child knows its smallest leaf nodeNum
	     so bubble sort them
	  */

	  for (i=0; i < numChildren; ++i)
	    for (j=i+1; j < numChildren; ++j)
	      if (T[i].tmp > T[j].tmp)
		swapChildren (i, j);

	  /*  now we know our smallest leaf nodeNum */

	  tmp = T[0].tmp;
	}
    }	


public int hashByLeaves (int[] hashValues, boolean isUnrooted, boolean doEdgeRemoval)
{
  if (isUnrooted)
    return hashByLeavesUnrooted (hashValues, doEdgeRemoval);
  else
    return hashByLeaves (hashValues);
}

// two constants with lots of bits, used in hashing.
// one of them is guaranteed to have a 32 element
// bit-rotation orbit (e.g. uniquely determined by the
// starting position of the four-1 run)
// HASH_CONST = 1000 0100 0110 1010 1110 1100 1111 0010 (binary)
//            =  8    4    6    a    e    c    f    2   (hex)

  final static int HASH_CONST = 0x846aecf2;

public final static int leafHash (int nodeNum)
{
  return (HASH_CONST << (nodeNum & 31)) | HASH_CONST >>> (32 - (nodeNum &  31)) * (1 + nodeNum);
}

  public int hashByLeaves (int[] hashValues)
    {
      // for each interior node N, compute the sum of (a + bx) over all leaf nodes x in the
      // tree rooted at N, where a and b are 32-bit constants
      // (any symmetric, associative hash function would do; i.e.  we just need invariance 
      // under permutations of the leaves)
      // this is to speed up the test for presence of a given partition of leaf nodes 
      // also, we record the hashValues in an array (in left-right-self order), using the mapping
      // from inode numbers (-1, -2, -3...) to (0, 1, 2 ...)

      int hv;

      if (numChildren == 0)
	{
	  hv = leafHash (nodeNum);
	}
      else
	{
	  // hash each child tree, then take the sum of their hash values

	  hv = 0;
	  for (int i=0; i < numChildren; ++i)
	    {
	      hv += T[i].hashByLeaves (hashValues);
	    }
	  hashValues[(-nodeNum)-1] = hv;
	}
      return hv;
    }

  public int hashByLeavesUnrooted (int[] hashValues, boolean edgeRemoval)
    {
      /*
       in an unrooted tree, we compute hash values as in hashByLeaves, from whatever
       the (arbitrary) root node is.
       Then simultaneously for each inode, we replace its hash value using a method
       depending on the value of edgeRemoval as follows:
       If edgeRemoval == false
         replace the original hash value with one of:
           a) the hash value of its left subtree 
           b) the hash value of its right subtree
           c) the hash value of the whole tree minus those for the left and right subtrees
         choosing the one which is computed on the largest number of nodes.
         This will allow us to quickly tell if a given 3-way split (induced by an inode) in the original
         tree is *not* present in a given randomized tree.
       else if edgeRemoval == true
         replace the original hash value with the larger of:
	   a) the original hash value
	   b) the hash value for the entire tree minus the original hash value
         this serves to hash the partition of leaf nodes into two sets obtained by deleting a tree
         edge.
       
       Hash values are recorded in hashValues as above.  If edgeRemoval==true, the hash value
       stored for an inode is the hash value for the partition formed by deleting the edge connecting
       that inode to its parent, and the hash value for the "root" is left unchanged.

      */

      // first, hash as if this is a rooted tree

      int hv = hashByLeaves (hashValues);

      // next, replace hash values with their modified forms:

      if (edgeRemoval)
	{
	  for (int i=0; i < hashValues.length; ++i)
	    hashValues[i] = (hashValues[i] >= hv - hashValues[i]) ? hashValues[i] : hv - hashValues[i];
	}
      else
	{
	  hashByLeavesUnrootedRecur (hv, hashValues);
	}
      return hv;
    };

public int hashByLeavesUnrootedRecur (int hvTotal, int[] hashValues)
{
  int oldHashValue;

  if (numChildren == 0)
    {
      // leaf node; compute its hash value
      oldHashValue = leafHash (nodeNum);
    }
  else
    {
      // save the original hash value, which we'll return to the caller

      oldHashValue = hashValues[(-nodeNum)-1];

      // find the largest of the original hash values for child trees

      int hv, hvSum;

      hv = hvSum = T[0].hashByLeavesUnrootedRecur (hvTotal, hashValues);
      for (int i=1; i < numChildren; ++i)
	{
	  int hvTmp = T[i].hashByLeavesUnrootedRecur (hvTotal, hashValues);
	  hvSum += hvTmp;
	  if (hvTmp > hv)
	    hv = hvTmp;
	}
      // use the hash value for the rest of the tree, if that is larger than
      // any of the hash values for subtrees; for non-"root" inodes, this corresponds
      // to the third "subtree" of the inode
      // for the "root" inode, hvTotal will equal hvSum, so the 
      // hash value will end up being the one for the largest of the three subtrees,
      // as required.

      hv  = (hv >= hvTotal - hvSum) ? hv : (hvTotal - hvSum);

      hashValues[(-nodeNum)-1] = hv;
    }
  return oldHashValue;
}

  public int getTreeExtents (int [] first, int [] last, int leafNum)
    {
      // for each interior node, record the first and last leaf nodes 
      // (in left-right depth-first traversal order)

      if (numChildren == 0)
	{
	  ++ leafNum;			// this is a leaf - bump up the counter
	}
      else
	{
	  first [(-nodeNum)-1] = leafNum;  // the next leaf visited will have this number
	  for (int i=0; i < numChildren; ++i)
	    leafNum = T[i].getTreeExtents (first, last, leafNum);
	  last [(-nodeNum)-1] = leafNum - 1;  // the last leaf visited had this number
	}
      return leafNum;
    }

  public int getLeafNodes (int [] nodeNums, int leafNum)
    {
      // fill the array nodeNums with the node numbers in the
      // order they are visited by left-right depth-first search

      if (numChildren == 0)
	nodeNums [leafNum++] = nodeNum;
      else
	for (int i = 0; i < numChildren; ++i)
	  leafNum = T[i].getLeafNodes (nodeNums, leafNum);
      return leafNum;
    }

  public void getInodes (ELTree [] inodes)
    {
      // fill an array with (references to) the subtrees of this tree
      //  using the mapping
      // from inode numbers (-1, -2, -3...) to array indicies (0, 1, 2 ...)  

      if (numChildren == 0)
	return;

      inodes[(-nodeNum)-1] = this;
      for (int i = 0; i < numChildren; ++i)
	T[i].getInodes (inodes);
    }

}   


